"""Simple 2D calibration
========================

.. autoclass:: Calibration2DSimple
   :members:
   :private-members:

"""

import numpy as np

from fluiddyn.io.image import imread


class Calibration2DSimple:
    """Simple 2D calibration

    The hypothesis are (at least):

    - No rotation (x and y remains x and y).
    - No optical aberration.
    - Everything orthogonal (the camera points directly toward the laser plan).

    Parameters
    ----------

    point0: sequence of size 2
      The coordinates in pixel of a first point as (indy, indx).

    point1: sequence of size 2
      The coordinates in pixel of a first point as (indy, indx).

    distance: number
      The distance between the 2 points in physical unit.

    aspect_ratio_pixel: number
      The aspect ratio of 1 pixel (a characteristic of the camera) as
      yphys1pxel/xphys1pxel.

    shape_image: sequence of size 2 (optional)
      The shape of the image. If it is not given, the next argument
      (`path_image`) is used.

    path_image: str (optional)
      The path toward the image (used to get the shape of the image).

    point_reference: sequence of size 2 (optional, default=(0, 0))
      A point for which (yphys, xphys) = orig_phys

    orig_phys: sequence of size 2 (optional, default=(0., 0.))
      The coordinates in physical unit of the point origin.

    """

    def __init__(
        self,
        point0,
        point1,
        distance,
        aspect_ratio_pixel=1.0,
        shape_image=None,
        path_image=None,
        point_origin=(0, 0),
        orig_phys=(0.0, 0.0),
    ):

        self.point0 = np.array(point0)
        self.point1 = np.array(point1)
        self.distance = distance

        delta_point = self.point1 - self.point0

        self.aspect_ratio_pixel = aspect_ratio_pixel

        self.xphys1pixel = distance / np.sqrt(
            delta_point[0] ** 2 + (aspect_ratio_pixel * delta_point[1]) ** 2
        )
        self.yphys1pixel = aspect_ratio_pixel * self.xphys1pixel

        if shape_image is None:
            arr = imread(path_image)
            shape_image = arr.shape
        self.shape_image = shape_image

        self.ix_origin = point_origin[1]
        self.iy_origin = point_origin[0]

        self.xphys_origin = orig_phys[1]
        self.yphys_origin = orig_phys[0]

    def pix2phys(self, ixs, iys):
        """Calcul physical coordinates from indices.

        Parameters
        ----------

        ixs: number or np.array
          Indices in the x direction (second dimension).

        iys: number or np.array
          Indices in the y direction (first dimension).

        """

        xphys = self.xphys_origin + self.xphys1pixel * (ixs - self.ix_origin)
        yphys = self.yphys_origin + self.yphys1pixel * (iys - self.iy_origin)
        return xphys, yphys

    def displ2phys(self, ixs, iys, displxs, displys):
        """Calcul physical coordinates and physical displacements from indices.

        Parameters
        ----------
        ixs: number or np.array
          Indices in the x direction (second dimension).

        iys: number or np.array
          Indices in the y direction (first dimension).

        displxs: number or np.array
          Displacements in the x direction (in pixels)

        displys: number or np.array
          Displacements in the y direction (in pixels)

        """
        xphys, yphys = self.pix2phys(ixs, iys)

        displ_phys_xs = self.xphys1pixel * displxs
        displ_phys_ys = self.yphys1pixel * displys

        return xphys, yphys, displ_phys_xs, displ_phys_ys



"""
Miscellaneous utilities
=======================

Provides:

.. autofunction:: imread

.. autofunction:: print_memory_usage

.. autofunction:: cstring

.. autofunction:: cprint

.. autofunction:: is_memory_full

.. autofunction:: str_short

"""

from pathlib import Path

import psutil
from IPython.lib.pretty import pretty

from fluiddyn.io.image import imread as _imread
from fluiddyn.io.image import imsave as _imsave
from fluiddyn.util import get_memory_usage
from fluiddyn.util import terminal_colors as term

color_dict = {
    "HEADER": term.HEADER,
    "OKBLUE": term.OKBLUE,
    "OKGREEN": term.OKGREEN,
    "WARNING": term.WARNING,
    "FAIL": term.FAIL,
    "ENDC": term.ENDC,
}


def imread(path):
    """Flatten image as a single gray-scale layer and
    loads as a numpy floating point array.

    """
    if isinstance(path, Path):
        path = str(path)
    # pylint: disable=W0703
    try:
        array = _imread(path)
    except Exception as error:
        raise type(error)(path).with_traceback(error.__traceback__)

    return array


def imsave(path, array, **kwargs):
    "tmp docstring"
    _imsave(path, array, **kwargs)


imsave.__doc__ = _imsave.__doc__


def _get_txt_memory_usage(string="Memory usage", color="OKGREEN"):
    mem = get_memory_usage()
    cstr = cstring((string + ": ").ljust(30) + f"{mem:.3f} Mb", color=color)
    return cstr


def print_memory_usage(string="Memory usage", color="OKGREEN"):
    """Print the memory usage."""
    print(_get_txt_memory_usage(string, color))


def cstring(*args, **kwargs):
    """Return a coloured string."""

    if "color" in kwargs:
        color = kwargs["color"]
    else:
        color = "OKBLUE"

    cstr = " ".join(args)
    return color_dict[color] + cstr + color_dict["ENDC"]


def cprint(*args, **kwargs):
    """Print with terminal colors."""

    cstr = cstring(*args, **kwargs)
    print(cstr)


def is_memory_full():
    """Checks if available system virtual memory is nearly saturated.

    Returns
    -------
    type: bool
        `True` if memory usage > 90 % and available memory < 500 MB.
        `False` otherwise
    """
    from .log import log_memory_usage

    mem = psutil.virtual_memory()

    if mem.percent > 90 or mem.available < 500 * 1024**2:
        log_memory_usage("Memory full! Current process using", mode="info")
        return True

    else:
        return False


def str_short(obj):
    """Give a short str for classes, function, etc."""

    try:
        return obj.__module__ + "." + obj.__name__
    except AttributeError:
        return pretty(obj)



"""
FluidImage
==========

"""

import os
import sys
from warnings import warn
from subprocess import getoutput
from pathlib import Path

if "OMP_NUM_THREADS" not in os.environ:
    if "numpy" in sys.modules:
        warn(
            "The environment variable OMP_NUM_THREADS "
            "was not set and numpy is already imported "
            'so fluidimage can not set OMP_NUM_THREADS to "1". '
            "It can be very bad for the performance of fluidimage topologies!"
        )
    else:
        warn(
            "The environment variable OMP_NUM_THREADS "
            'was not set so fluidimage fixes it to "1", '
            "which is needed for fluidimage topologies."
        )
        os.environ["OMP_NUM_THREADS"] = "1"

import numpy as np

from fluiddyn.util import create_object_from_file, get_memory_usage
from fluiddyn.util.paramcontainer import ParamContainer
from fluiddyn.util.serieofarrays import SerieOfArraysFromFiles, SeriesOfArrays
from fluidimage.topologies.log import LogTopology

from ._version import __version__
from .util import (
    config_logging,
    imread,
    imsave,
    log_memory_usage,
    logger,
    print_memory_usage,
    reset_logger,
)


if any(
    any(test_tool in arg for arg in sys.argv)
    for test_tool in ("pytest", "unittest")
):
    print(
        "Fluidimage guesses that it is tested so it"
        " loads the Agg Matplotlib backend."
    )
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    plt.show = lambda: None


def get_path_image_samples():
    try:
        from ._path_image_samples import path_image_samples
    except ImportError:
        pass
    else:
        if path_image_samples.exists():
            return path_image_samples

    path_image_samples = (
        Path.home() / ".local/fluidimage/repository/image_samples"
    )
    path_repo = path_image_samples.parent
    path_repo.mkdir(parents=True, exist_ok=True)
    if not path_image_samples.exists():
        getoutput(
            f"hg clone https://foss.heptapod.net/fluiddyn/fluidimage {path_repo}"
        )
    return path_image_samples


__all__ = [
    "__version__",
    "create_object_from_file",
    "get_memory_usage",
    "imread",
    "imsave",
    "logger",
    "config_logging",
    "reset_logger",
    "log_memory_usage",
    "print_memory_usage",
    "ParamContainer",
    "LogTopology",
    "get_path_image_samples",
    "SeriesOfArrays",
    "SerieOfArraysFromFiles",
    "np",
]



"""Topology for surface tracking(:mod:`fluidimage.topologies.surface_tracking`)
===============================================================================

.. autoclass:: TopologySurfaceTracking
   :members:
   :private-members:

.. todo::

   Make this code usable:

   - good unittests
   - tutorial with ipynb
   - example integrated in the documentation

"""
import json
import sys
from pathlib import Path

from fluiddyn.io.image import imsave_h5
from fluidimage import ParamContainer, SerieOfArraysFromFiles, SeriesOfArrays
from fluidimage.topologies import prepare_path_dir_result
from fluidimage.util import imread, logger
from fluidimage.works.surface_tracking import WorkSurfaceTracking

from . import image2image
from .base import TopologyBase


class TopologySurfaceTracking(TopologyBase):
    """Topology for SurfaceTracking.

    Parameters
    ----------

    params : None

      A ParamContainer containing the parameters for the computation.

    logging_level : str, {'warning', 'info', 'debug', ...}

      Logging level.

    nb_max_workers : None, int

      Maximum numbers of "workers". If None, a number is computed from the
      number of cores detected. If there are memory errors, you can try to
      decrease the number of workers.

    """

    @classmethod
    def create_default_params(cls):
        """Class method returning the default parameters.

        For developers: cf. fluidsim.base.params

        """
        params = ParamContainer(tag="params")

        WorkSurfaceTracking._complete_params_with_default(params)

        params._set_child(
            "images",
            attribs={
                "path": "",
                "path_ref": "",
                "str_slice_ref": None,
                "str_slice": None,
            },
        )

        params.images._set_doc(
            """
Parameters indicating the input image set.

path : str, {''}

    String indicating the input images (can be a full path towards an image
    file or a string given to `glob`).

path_ref : str, {''}

    String indicating the reference input images (can be a full path towards
    an image file or a string given to `glob`).

str_slice_ref : None

    String indicating as a Python slicing how to select reference images
    from the serie of reference images on the disk (in order to compute
    k_x value necessary for gain and filter).
    If None, no selection so all images are going to be processed.

str_slice : None

    String indicating as a Python slicing how to select images from the
    serie of images on the disk. If None, no selection so all images
    are going to be processed.

"""
        )

        params._set_child(
            "saving", attribs={"path": None, "how": "ask", "postfix": "pre"}
        )

        params.saving._set_doc(
            """Saving of the results.

path : None or str

    Path of the directory where the data will be saved. If None, the path is
    obtained from the input path and the parameter `postfix`.

how : str {'ask'}

    'ask', 'new_dir', 'complete' or 'recompute'.

postfix : str

    Postfix from which the output file is computed.
"""
        )

        params._set_internal_attr(
            "_value_text",
            json.dumps(
                {
                    "program": "fluidimage",
                    "module": "fluidimage.topologies.surface_tracking",
                    "class": "TopologySurfaceTracking",
                }
            ),
        )

        params._set_child("preproc")
        image2image.complete_im2im_params_with_default(params.preproc)

        return params

    def __init__(self, params, logging_level="info", nb_max_workers=None):

        self.params = params

        if params.surface_tracking is None:
            raise ValueError("params.surface_tracking has to be set.")

        self.serie = SerieOfArraysFromFiles(
            params.images.path, params.images.str_slice
        )
        self.series = SeriesOfArrays(
            params.images.path,
            "i:i+"
            + str(self.serie.get_index_slices()[0][2] + 1)
            + ":"
            + str(self.serie.get_index_slices()[0][2]),
            ind_start=self.serie.get_index_slices()[0][0],
            ind_stop=self.serie.get_index_slices()[0][1] - 1,
            ind_step=self.serie.get_index_slices()[0][2],
        )
        path_dir = self.serie.path_dir
        path_dir_result, self.how_saving = prepare_path_dir_result(
            path_dir, params.saving.path, params.saving.postfix, params.saving.how
        )

        self.path_dir_result = path_dir_result
        self.path_dir_src = Path(path_dir)

        self.surface_tracking_work = WorkSurfaceTracking(params)

        super().__init__(
            path_dir_result=path_dir_result,
            logging_level=logging_level,
            nb_max_workers=nb_max_workers,
        )

        queue_paths = self.add_queue("paths")
        queue_couples_of_names = self.add_queue("couples of names")
        queue_arrays = queue_arrays1 = self.add_queue("arrays")
        queue_angles = self.add_queue("angles")
        queue_couples_of_arrays = self.add_queue(
            "couples of corrected angles and angles"
        )
        queuemod0_angles = self.add_queue("corrected angles copy")
        queuemod_angles = self.add_queue("corrected angles")
        queue_heights_and_shapes = self.add_queue("heights and shapes")
        queue_heights = self.add_queue("heights")

        if params.preproc.im2im is not None:
            queue_arrays1 = self.add_queue("arrays1")

        self.add_work(
            "fill_path",
            self.fill_queue_paths,
            output_queue=(queue_paths, queue_couples_of_names),
            kind="one shot",
        )

        self.add_work(
            "read_array",
            self.imread,
            input_queue=queue_paths,
            output_queue=queue_arrays,
            kind="io",
        )

        if params.preproc.im2im is not None:
            im2im_func = image2image.TopologyImage2Image.init_im2im(
                self, params.preproc
            )

            self.add_work(
                "image2image",
                func_or_cls=im2im_func,
                input_queue=queue_arrays,
                output_queue=queue_arrays1,
            )

        self.add_work(
            "process_frame",
            self.surface_tracking_work.process_frame_func,
            input_queue=queue_arrays1,
            output_queue=queue_angles,
        )

        self.add_work(
            "create_couple",
            self.make_couples,
            input_queue=(queuemod0_angles, queue_angles, queue_couples_of_names),
            output_queue=(queuemod_angles, queue_couples_of_arrays),
            kind="global",
        )

        self.add_work(
            "correct_couple_of_phases",
            self.surface_tracking_work.correctcouple,
            input_queue=queue_couples_of_arrays,
            output_queue=queuemod0_angles,
        )

        self.add_work(
            "calcul_height",
            self.surface_tracking_work.calculheight_func,
            input_queue=queuemod_angles,
            output_queue=queue_heights_and_shapes,
        )

        self.add_work(
            "set_borders_zero",
            self.surface_tracking_work.set_borders_zero_func,
            input_queue=queue_heights_and_shapes,
            output_queue=queue_heights,
        )

        self.add_work(
            "save", self.save_image, input_queue=queue_heights, kind="io"
        )

    def make_couples(self, input_queues, output_queues):
        """correctphase"""
        queue_couples_of_names = input_queues[2]
        queue_angles = input_queues[1]
        queuemod0_angles = input_queues[0]
        queue_couple = output_queues[1]
        if not (queue_couples_of_names):
            for key in tuple(queuemod0_angles):
                output_queues[0][key] = queuemod0_angles[key]
                del queuemod0_angles[key]
        if not (queue_angles):
            print("no queue")
            return
        for key, couple in tuple(queue_couples_of_names.items()):
            # if correspondant arrays are available, make an array couple
            if couple[0] is couple[1]:
                if couple[0] in queue_angles.keys():
                    array1 = queue_angles[couple[0]]
                    array2 = queue_angles[couple[0]]

                    queue_couple[couple[0]] = (array1, array2)
                    del queue_angles[couple[0]]
                    del queue_couples_of_names[key]
            elif (
                couple[0] in queuemod0_angles.keys()
                and couple[1] in queue_angles.keys()
            ):
                array1 = queuemod0_angles[couple[0]]
                array2 = queue_angles[couple[1]]

                queue_couple[couple[1]] = (array1, array2)
                del queue_angles[couple[1]]
                del queue_couples_of_names[key]
                output_queues[0][couple[0]] = queuemod0_angles[couple[0]]
                del queuemod0_angles[couple[0]]

    def imread(self, path):
        array = imread(path)
        return (array, path)

    def save_image(self, tuple_image_path):
        image, path = tuple_image_path
        name_file = Path(path).name
        path_out = self.path_dir_result / name_file
        imsave_h5(path_out, image, splitext=False)

    def fill_queue_paths(self, input_queue, output_queues):

        assert input_queue is None
        queue_paths = output_queues[0]
        queue_couples_of_names = output_queues[1]

        serie = self.serie
        if len(serie) == 0:
            logger.warning("add 0 image. No image to process.")
            return

        names = serie.get_name_arrays()
        for name in names:
            path_im_output = self.path_dir_result / name
            path_im_input = str(self.path_dir_src / name)
            if self.how_saving == "complete":
                if not path_im_output.exists():
                    queue_paths[name] = path_im_input
            else:
                queue_paths[name] = path_im_input

        if len(names) == 0:
            if self.how_saving == "complete":
                logger.warning(
                    'topology in mode "complete" and work already done.'
                )
            else:
                logger.warning("Nothing to do")
            return

        nb_names = len(names)
        logger.info(f"Add {nb_names} images to compute.")
        logger.info("First files to process: " + str(names[:4]))

        logger.debug("All files: " + str(names))

        series = self.series
        if not series:
            logger.warning("add 0 couple. No phase to correct.")
            return

        nb_series = len(series)
        logger.info(f"Add {nb_series} phase to correct.")

        for iserie, serie in enumerate(series):
            if iserie > 1:
                break
            logger.info(
                "Files of serie {}: {}".format(iserie, serie.get_name_arrays())
            )
        # for the first corrected angle : corrected_angle = angle
        ind_serie, serie = next(series.items())
        name = serie.get_name_arrays()[0]
        queue_couples_of_names[ind_serie - 1] = (name, name)
        for ind_serie, serie in series.items():
            queue_couples_of_names[ind_serie] = serie.get_name_arrays()


if "sphinx" in sys.modules:
    params = TopologySurfaceTracking.create_default_params()

    __doc__ += params._get_formatted_docs()



import matplotlib.pyplot as plt
import numpy as np
from numpy import pi

from fluidimage.calcul.interpolate.thin_plate_spline import (
    compute_tps_coeff_iter,
    compute_tps_matrices_dxy,
    compute_tps_matrix,
)


def myplot(i, x, y, U):
    plt.figure(i)
    ax = plt.gca()

    ax.scatter(x, y, c=U, vmin=-1, vmax=1)


# set of random x coordinates from 0 to 2pi
x = 2 * np.pi * np.random.rand(100)
y = 2 * np.pi * np.random.rand(100)

U = np.exp(-((x - pi) ** 2 + (y - pi) ** 2))  # gaussian
# U = x  # linear

myplot(0, x, y, U)


# calculate tps coeff
centers = np.vstack([x, y])
smoothing_coef = 0
subdom_size = 20
U_smooth, U_tps = compute_tps_coeff_iter(centers, U, smoothing_coef, threshold=1)

# interpolation grid
xI = yI = np.arange(0, 2 * pi, 0.1)
XI, YI = np.meshgrid(xI, yI)
XI = XI.ravel()
YI = YI.ravel()

new_positions = np.vstack([XI, YI])

# evaluate interpolation on the new grid
EM = compute_tps_matrix(new_positions, centers)

U_eval = np.dot(U_tps, EM)

myplot(1, XI, YI, U_eval)

DMX, DMY = compute_tps_matrices_dxy(new_positions, centers)
DUX_eval = np.dot(U_tps, DMX)
DUY_eval = np.dot(U_tps, DMY)

myplot(2, XI, YI, DUX_eval)
myplot(3, XI, YI, DUY_eval)

plt.show()



"""Topology for PIV computation (:mod:`fluidimage.topologies.piv`)
==================================================================

.. autoclass:: TopologyPIV
   :members:
   :private-members:

"""
import copy
import json
import sys

from fluidimage import ParamContainer, SeriesOfArrays
from fluidimage.data_objects.piv import ArrayCouple, get_name_piv
from fluidimage.topologies import TopologyBase, prepare_path_dir_result
from fluidimage.util import DEBUG, imread, logger
from fluidimage.works.piv import WorkPIV

from . import image2image


def is_name_in_queue(image_name, queue):
    """Check if a name is in a queue of series"""
    for names in queue.values():
        if image_name in names:
            return True
    return False


class TopologyPIV(TopologyBase):
    """Topology for PIV computation.

    The most useful methods for the user (in particular :func:`compute`) are
    defined in the base class :class:`fluidimage.topologies.base.TopologyBase`.

    Parameters
    ----------

    params : None

      A ParamContainer (created with the class method
      :func:`create_default_params`) containing the parameters for the
      computation.

    logging_level : str, {'warning', 'info', 'debug', ...}

      Logging level.

    nb_max_workers : None, int

      Maximum numbers of "workers". If None, a number is estimated from the
      number of cores detected. If there are memory errors, you can try to
      decrease the number of workers.

    """

    WorkVelocimetry = WorkPIV

    @classmethod
    def create_default_params(cls):
        """Class method returning the default parameters.

        Typical usage::

          params = TopologyPIV.create_default_params()
          # modify parameters here
          ...

          topo = TopologyPIV(params)

        """
        params = ParamContainer(tag="params")

        params._set_child(
            "series",
            attribs={
                "path": "",
                "strcouple": "i:i+2",
                "ind_start": 0,
                "ind_stop": None,
                "ind_step": 1,
            },
        )

        params.series._set_doc(
            """
Parameters indicating the input series of images.

path : str, {''}

    String indicating the input images (can be a full path towards an image
    file or a string given to `glob`).

strcouple : 'i:i+2'

    String indicating as a Python slicing how couples of images are formed.
    There is one couple per value of `i`. The values of `i` are set with the
    other parameters `ind_start`, `ind_step` and `ind_stop` approximately with
    the function range (`range(ind_start, ind_stop, ind_step)`).

    Python slicing is a very powerful notation to define subset from a
    (possibly multidimensional) set of images. For a user, an alternative is to
    understand how Python slicing works. See for example this page:
    http://stackoverflow.com/questions/509211/explain-pythons-slice-notation.

    Another possibility is to follow simple examples:

    For single-frame images (im0, im1, im2, im3, ...), we keep the default
    value 'i:i+2' to form the couples (im0, im1), (im1, im2), ...

    To see what it gives, one can use IPython and range:

    >>> i = 0
    >>> list(range(10))[i:i+2]
    [0, 1]

    >>> list(range(10))[i:i+4:2]
    [0, 2]

    We see that we can also use the value 'i:i+4:2' to form the couples (im0,
    im2), (im1, im3), ...

    For double-frame images (im1a, im1b, im2a, im2b, ...) you can write

    >>> params.series.strcouple = 'i, 0:2'

    In this case, the first couple will be (im1a, im1b).

    To get the first couple (im1a, im1a), we would have to write

    >>> params.series.strcouple = 'i:i+2, 0'

ind_start : int, {0}

ind_step : int, {1}

int_stop : None

"""
        )

        params._set_child(
            "saving", attribs={"path": None, "how": "ask", "postfix": "piv"}
        )

        params.saving._set_doc(
            """Saving of the results.

path : None or str

    Path of the directory where the data will be saved. If None, the path is
    obtained from the input path and the parameter `postfix`.

how : str {'ask'}

    'ask', 'new_dir', 'complete' or 'recompute'.

postfix : str

    Postfix from which the output file is computed.
"""
        )

        cls.WorkVelocimetry._complete_params_with_default(params)

        params._set_internal_attr(
            "_value_text",
            json.dumps(
                {
                    "program": "fluidimage",
                    "module": "fluidimage.topologies.piv",
                    "class": "TopologyPIV",
                }
            ),
        )

        params._set_child("preproc")
        image2image.complete_im2im_params_with_default(params.preproc)

        return params

    def __init__(self, params, logging_level="info", nb_max_workers=None):

        self.params = params

        self.series = SeriesOfArrays(
            params.series.path,
            params.series.strcouple,
            ind_start=params.series.ind_start,
            ind_stop=params.series.ind_stop,
            ind_step=params.series.ind_step,
        )

        path_dir = self.series.serie.path_dir
        path_dir_result, self.how_saving = prepare_path_dir_result(
            path_dir, params.saving.path, params.saving.postfix, params.saving.how
        )

        super().__init__(
            path_dir_result=path_dir_result,
            logging_level=logging_level,
            nb_max_workers=nb_max_workers,
        )

        queue_couples_of_names = self.add_queue("couples of names")
        queue_paths = self.add_queue("paths")
        queue_arrays = queue_arrays1 = self.add_queue("arrays")
        queue_couples_of_arrays = self.add_queue("couples of arrays")
        queue_piv = self.add_queue("piv")

        if params.preproc.im2im is not None:
            queue_arrays1 = self.add_queue("arrays1")

        self.add_work(
            "fill (couples of names, paths)",
            func_or_cls=self.fill_couples_of_names_and_paths,
            output_queue=(queue_couples_of_names, queue_paths),
            kind=("global", "one shot"),
        )
        self.add_work(
            "read array",
            func_or_cls=imread,
            input_queue=queue_paths,
            output_queue=queue_arrays,
            kind="io",
        )

        if params.preproc.im2im is not None:
            im2im_func = image2image.TopologyImage2Image.init_im2im(
                self, params.preproc
            )

            self.add_work(
                "image2image",
                func_or_cls=im2im_func,
                input_queue=queue_arrays,
                output_queue=queue_arrays1,
            )

        self.add_work(
            "make couples of arrays",
            func_or_cls=self.make_couples,
            params_cls=None,
            input_queue=(queue_couples_of_names, queue_arrays1),
            output_queue=queue_couples_of_arrays,
            kind="global",
        )

        self.work_piv = self.WorkVelocimetry(self.params)

        self.add_work(
            "compute piv",
            func_or_cls=self.work_piv.calcul,
            params_cls=params,
            input_queue=queue_couples_of_arrays,
            output_queue=queue_piv,
        )

        self.add_work(
            "save piv",
            func_or_cls=self.save_piv_object,
            input_queue=queue_piv,
            kind="io",
        )
        self.results = []

    def save_piv_object(self, obj):
        """Save a PIV object"""
        ret = obj.save(self.path_dir_result)
        self.results.append(ret)

    def fill_couples_of_names_and_paths(self, input_queue, output_queues):
        """Fill the two first queues"""
        assert input_queue is None
        queue_couples_of_names = output_queues[0]
        queue_paths = output_queues[1]

        series = self.series
        if not series:
            logger.warning("add 0 couple. No PIV to compute.")
            return
        if self.how_saving == "complete":
            index_series = []
            for ind_serie, serie in self.series.items():
                name_piv = get_name_piv(serie, prefix="piv")
                if not (self.path_dir_result / name_piv).exists():
                    index_series.append(ind_serie)

            if not index_series:
                logger.warning(
                    'topology in mode "complete" and work already done.'
                )
                return

            series.set_index_series(index_series)

            if logger.isEnabledFor(DEBUG):
                logger.debug(repr([serie.get_name_arrays() for serie in series]))

        nb_series = len(series)
        logger.info(f"Add {nb_series} PIV fields to compute.")

        for iserie, serie in enumerate(series):
            if iserie > 1:
                break
            logger.info(
                "Files of serie {}: {}".format(iserie, serie.get_name_arrays())
            )

        for ind_serie, serie in series.items():
            queue_couples_of_names[ind_serie] = serie.get_name_arrays()
            for name, path in serie.get_name_path_arrays():
                queue_paths[name] = path

    def make_couples(self, input_queues, output_queue):
        """Make the couples of arrays"""
        queue_couples_of_names = input_queues[0]
        queue_arrays = input_queues[1]

        try:
            params_mask = self.params.mask
        except AttributeError:
            params_mask = None
        # for each name couple
        for key, couple in tuple(queue_couples_of_names.items()):
            # if correspondant arrays are available, make an array couple
            if (
                couple[0] in queue_arrays.keys()
                and couple[1] in queue_arrays.keys()
            ):
                array1 = queue_arrays[couple[0]]
                array2 = queue_arrays[couple[1]]
                serie = copy.copy(self.series.get_serie_from_index(key))

                # logger.debug(
                #     f"create couple {key}: {couple}, ({array1}, {array2})"
                # )
                array_couple = ArrayCouple(
                    names=(couple[0], couple[1]),
                    arrays=(array1, array2),
                    params_mask=params_mask,
                    serie=serie,
                )
                output_queue[key] = array_couple
                del queue_couples_of_names[key]
                # remove the image_array if it not will be used anymore
                if not is_name_in_queue(couple[0], queue_couples_of_names):
                    del queue_arrays[couple[0]]
                if not is_name_in_queue(couple[1], queue_couples_of_names):
                    del queue_arrays[couple[1]]

    def make_text_at_exit(self, time_since_start):
        """Make a text printed at exit"""

        txt = f"Stop compute after t = {time_since_start:.2f} s"
        try:
            nb_results = len(self.results)
        except AttributeError:
            nb_results = None
        if nb_results is not None and nb_results > 0:
            txt += f" ({nb_results} piv fields, {time_since_start / nb_results:.2f} s/field)."
        else:
            txt += "."

        txt += "\npath results:\n" + str(self.path_dir_result)

        return txt


Topology = TopologyPIV

if "sphinx" in sys.modules:
    params = Topology.create_default_params()
    __doc__ += params._get_formatted_docs()


if __name__ == "__main__":
    from fluidimage import get_path_image_samples

    params = Topology.create_default_params()

    params.series.path = str(get_path_image_samples() / "Karman/Images")
    params.series.ind_start = 1
    params.series.ind_step = 2

    params.piv0.shape_crop_im0 = 32
    params.multipass.number = 2
    params.multipass.use_tps = False

    params.mask.strcrop = ":, 50:500"

    params.preproc.im2im = "numpy.ones_like"

    # params.saving.how = 'complete'
    params.saving.postfix = "piv_example"

    topo = Topology(params, logging_level="info")

    topo.make_code_graphviz("tmp.dot")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mainwindow.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

# !!! Changed by hand to get the tests working !!!
# before it was `from PyQt5 ...`
from matplotlib.backends.qt_compat import QtCore, QtGui, QtWidgets


class Ui_MainWindow:
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(775, 901)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 775, 24))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuTopologies = QtWidgets.QMenu(self.menubar)
        self.menuTopologies.setObjectName("menuTopologies")
        self.menuComputations = QtWidgets.QMenu(self.menubar)
        self.menuComputations.setObjectName("menuComputations")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionOpen = QtWidgets.QAction(MainWindow)
        self.actionOpen.setObjectName("actionOpen")
        self.actionTopologyPIV = QtWidgets.QAction(MainWindow)
        self.actionTopologyPIV.setObjectName("actionTopologyPIV")
        self.actionTopologyPreproc = QtWidgets.QAction(MainWindow)
        self.actionTopologyPreproc.setObjectName("actionTopologyPreproc")
        self.menuFile.addAction(self.actionOpen)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuTopologies.menuAction())
        self.menubar.addAction(self.menuComputations.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuTopologies.setTitle(_translate("MainWindow", "Topologies"))
        self.menuComputations.setTitle(_translate("MainWindow", "Computations"))
        self.actionOpen.setText(_translate("MainWindow", "Open"))
        self.actionTopologyPIV.setText(_translate("MainWindow", "TopologyPIV"))
        self.actionTopologyPreproc.setText(
            _translate("MainWindow", "TopologyPreproc")
        )



"""Topology for BOS computation (:mod:`fluidimage.topologies.bos`)
==================================================================

.. autoclass:: TopologyBOS
   :members:
   :private-members:

"""
import json
import sys
from pathlib import Path

from fluidimage import ParamContainer, SerieOfArraysFromFiles
from fluidimage.data_objects.piv import ArrayCoupleBOS, get_name_bos
from fluidimage.topologies import TopologyBase, prepare_path_dir_result
from fluidimage.util import imread, logger
from fluidimage.works.piv import WorkPIV

from . import image2image


class TopologyBOS(TopologyBase):
    """Topology for BOS computation.

    See https://en.wikipedia.org/wiki/Background-oriented_schlieren_technique

    The most useful methods for the user (in particular :func:`compute`) are
    defined in the base class :class:`fluidimage.topologies.base.TopologyBase`.

    Parameters
    ----------

    params : None

      A ParamContainer (created with the class method
      :func:`create_default_params`) containing the parameters for the
      computation.

    logging_level : str, {'warning', 'info', 'debug', ...}

      Logging level.

    nb_max_workers : None, int

      Maximum numbers of "workers". If None, a number is computed from the number of
      cores detected. If there are memory errors, you can try to decrease the number
      of workers.

    """

    @classmethod
    def create_default_params(cls):
        """Class method returning the default parameters.

        Typical usage::

          params = TopologyPIV.create_default_params()
          # modify parameters here
          ...

          topo = TopologyPIV(params)

        """
        params = ParamContainer(tag="params")

        params._set_child("images", attribs={"path": "", "str_slice": None})

        params.images._set_doc(
            """
Parameters indicating the input image set.

path : str, {''}

    String indicating the input images (can be a full path towards an image
    file or a string given to `glob`).

str_slice : None

    String indicating as a Python slicing how to select images from the serie of
    images on the disk. If None, no selection so all images are going to be
    processed.

"""
        )

        params._set_attrib("reference", 0)

        params._set_doc(
            """
reference : str or int, {0}

    Reference file (from which the displacements will be computed). Can be an
    absolute file path, a file name or the index in the list of files found
    from the parameters in ``params.images``.

"""
        )

        params._set_child(
            "saving", attribs={"path": None, "how": "ask", "postfix": "bos"}
        )

        params.saving._set_doc(
            """Saving of the results.

path : None or str

    Path of the directory where the data will be saved. If None, the path is
    obtained from the input path and the parameter `postfix`.

how : str {'ask'}

    'ask', 'new_dir', 'complete' or 'recompute'.

postfix : str

    Postfix from which the output file is computed.
"""
        )

        WorkPIV._complete_params_with_default(params)

        params._set_internal_attr(
            "_value_text",
            json.dumps(
                {
                    "program": "fluidimage",
                    "module": "fluidimage.topologies.bos",
                    "class": "TopologyBOS",
                }
            ),
        )

        params._set_child("preproc")
        image2image.complete_im2im_params_with_default(params.preproc)

        return params

    def __init__(self, params, logging_level="info", nb_max_workers=None):

        self.params = params

        self.serie = SerieOfArraysFromFiles(
            params.images.path, params.images.str_slice
        )

        path_dir = Path(self.serie.path_dir)
        path_dir_result, self.how_saving = prepare_path_dir_result(
            path_dir, params.saving.path, params.saving.postfix, params.saving.how
        )

        self.path_dir_result = path_dir_result
        self.path_dir_src = Path(path_dir)

        if not isinstance(params.reference, int):
            reference = Path(params.reference).expanduser()
        else:
            reference = params.reference

        if isinstance(reference, int):
            names = self.serie.get_name_arrays()
            names = sorted(names)
            path_reference = path_dir / names[reference]

        else:
            reference = Path(reference)
            if reference.is_file():
                path_reference = reference
            else:
                path_reference = path_dir_result / reference
                if not path_reference.is_file():
                    raise ValueError(
                        "Bad value of params.reference:" + path_reference
                    )

        self.name_reference = path_reference.name
        self.path_reference = path_reference
        self.image_reference = imread(path_reference)

        super().__init__(
            path_dir_result=path_dir_result,
            logging_level=logging_level,
            nb_max_workers=nb_max_workers,
        )

        queue_paths = self.add_queue("paths")
        queue_arrays = queue_arrays1 = self.add_queue("arrays")
        queue_bos = self.add_queue("bos")

        if params.preproc.im2im is not None:
            queue_arrays1 = self.add_queue("arrays1")

        self.add_work(
            "fill paths",
            func_or_cls=self.fill_queue_paths,
            output_queue=queue_paths,
            kind=("global", "one shot"),
        )

        def _imread(path):
            image = imread(path)
            return image, Path(path).name

        self.add_work(
            "read array",
            func_or_cls=_imread,
            input_queue=queue_paths,
            output_queue=queue_arrays,
            kind="io",
        )

        if params.preproc.im2im is not None:
            im2im_func = image2image.TopologyImage2Image.init_im2im(
                self, params.preproc
            )

            self.add_work(
                "image2image",
                func_or_cls=im2im_func,
                input_queue=queue_arrays,
                output_queue=queue_arrays1,
            )

        self.add_work(
            "compute bos",
            func_or_cls=self.calcul,
            params_cls=params,
            input_queue=queue_arrays,
            output_queue=queue_bos,
        )

        self.add_work(
            "save bos",
            func_or_cls=self.save_bos_object,
            input_queue=queue_bos,
            kind="io",
        )

    def save_bos_object(self, obj):
        """Save a BOS object"""
        ret = obj.save(self.path_dir_result, kind="bos")
        return ret

    def calcul(self, tuple_image_path):
        """Compute a BOS field"""
        image, name = tuple_image_path
        array_couple = ArrayCoupleBOS(
            names=(self.name_reference, name),
            arrays=(self.image_reference, image),
            params_mask=self.params.mask,
            serie=self.serie,
            paths=[self.path_reference, self.path_dir_src / name],
        )
        return WorkPIV(self.params).calcul(array_couple)

    def fill_queue_paths(self, input_queue, output_queue):
        """Fill the first queue (paths)"""
        assert input_queue is None

        serie = self.serie
        if not serie:
            logger.warning("add 0 image. No image to process.")
            return

        names = serie.get_name_arrays()

        for name in names:
            path_im_output = self.path_dir_result / get_name_bos(name, serie)
            path_im_input = str(self.path_dir_src / name)
            if self.how_saving == "complete":
                if not path_im_output.exists():
                    output_queue[name] = path_im_input
            else:
                output_queue[name] = path_im_input

        if not names:
            if self.how_saving == "complete":
                logger.warning(
                    'topology in mode "complete" and work already done.'
                )
            else:
                logger.warning("Nothing to do")
            return

        try:
            del output_queue[self.path_reference]
        except KeyError:
            pass

        nb_names = len(names)
        logger.info(f"Add {nb_names} images to compute.")
        logger.info(f"First files to process: {names[:4]}")

        logger.debug(f"All files: {names}")

    def make_text_at_exit(self, time_since_start):
        """Make a text printed at exit"""
        txt = f"Stop compute after t = {time_since_start:.2f} s"
        try:
            nb_results = len(self.results)
        except AttributeError:
            nb_results = None
        if nb_results is not None and nb_results > 0:
            txt += f" ({nb_results} bos fields, {time_since_start / nb_results:.2f} s/field)."
        else:
            txt += "."

        txt += f"\npath results:\n{self.path_dir_result}"

        return txt


if "sphinx" in sys.modules:
    params = TopologyBOS.create_default_params()

    __doc__ += params._get_formatted_docs()

if __name__ == "__main__":
    params = TopologyBOS.create_default_params()
    params.series.path = "../../../image_samples/Karman/Images"
    params.series.ind_start = 1
    params.series.ind_step = 2

    params.piv0.shape_crop_im0 = 32
    params.multipass.number = 2
    params.multipass.use_tps = False

    params.mask.strcrop = ":, 50:500"

    # params.saving.how = 'complete'
    params.saving.postfix = "bos_example2"

    topo = TopologyBOS(params, logging_level="info")

    topo.make_code_graphviz("tmp.dot")



"""Preprocessing of images
==========================

Images acquired from experiments can suffer from noise, reflection and other
undesirable effects. Preprocessing images by careful use of thresholds and
filters can significantly improve the quality of PIV.

Provides:

.. autosummary::
   :toctree:

   base
   toolbox
   image2image

"""

from .base import PreprocBase

__all__ = ["PreprocBase"]



"""
Launchers for topologies used for GUI (:mod:`fluidimage.topologies.launcher`)
=============================================================================

.. autoclass:: TopologyLauncher
   :members:
   :private-members:

.. autoclass:: TopologyPIVLauncher
   :members:
   :private-members:

.. autoclass:: TopologyPreprocLauncher
   :members:
   :private-members:

"""
import inspect
import json

from fluidimage.topologies.piv import TopologyPIV
from fluidimage.topologies.preproc import TopologyPreproc


def _get_args_bound_method(method):
    try:  # python 3
        arg_spec = inspect.getfullargspec(method)
    except AttributeError:  # python 2
        arg_spec = inspect.getargspec(method.__func__)

    # key args without 'self' and 'params'
    key_args = arg_spec.args[2:]
    defaults = arg_spec.defaults[1:]
    return key_args, defaults


class TopologyLauncher:
    """Launcher (class to be subclassed)."""

    Topology = TopologyPIV

    @classmethod
    def create_default_params(cls):
        params = cls.Topology.create_default_params()

        key_args, defaults = _get_args_bound_method(cls.Topology.__init__)

        params._set_child("topology")

        for key, value in zip(key_args, defaults):
            params.topology._set_attrib(key, value)

        params.topology._set_doc(cls.Topology.__doc__)

        key_args, defaults = _get_args_bound_method(cls.Topology.compute)
        params.topology._set_child("compute")

        for key, value in zip(key_args, defaults):
            params.topology.compute._set_attrib(key, value)

        params.topology.compute._set_doc(cls.Topology.compute.__doc__)

        params._set_internal_attr(
            "_value_text",
            json.dumps(
                {
                    "program": "fluidimage",
                    "module": cls.__module__,
                    "class": cls.__name__,
                }
            ),
        )

        return params

    def __init__(self, params):
        self.params = params
        kwargs = params.topology._make_dict_attribs()
        print(kwargs)
        self.topology = self.Topology(params, **kwargs)

    def compute(self):
        kwargs = self.params.topology.compute._make_dict_attribs()
        print(kwargs)
        self.topology.compute(**kwargs)


class TopologyPIVLauncher(TopologyLauncher):
    """PIV topology launcher"""


class TopologyPreprocLauncher(TopologyLauncher):
    """Preproc topology launcher"""

    Topology = TopologyPreproc



"""Handles I/O of images to tools(:mod:`fluidimage.preproc.io`)
===============================================================

Provides decorators to handle input and output of images for tools/functions in
:mod:`fluidimage.preproc.toolbox`.

"""

import numpy as np
from decorator import decorator


def _get_img_arg(*args, **kwargs):
    if "img" in kwargs:
        return kwargs["img"]

    else:
        return args[0]


def _replace_img_arg(new_img, *args, **kwargs):
    if not isinstance(args, list):
        args = list(args)

    if "img" in kwargs:
        kwargs["img"] = new_img
    else:
        args[0] = new_img

    return args, kwargs


def _get_array_like_type(array_like, convert_to_ndarray=False):
    if not isinstance(array_like, np.ndarray):
        ArrayLike = array_like.__class__
    else:
        ArrayLike = np.array

    if convert_to_ndarray:
        return np.array(array_like), ArrayLike

    else:
        return ArrayLike


@decorator
def iterate_multiple_imgs(tool, *args, **kwargs):
    """
    Feeds one image at a time to the function `tool`,
    typically a spatial filter, or a brightness/contrast adjustment tool.

    """
    img_array_in = _get_img_arg(*args, **kwargs)

    if isinstance(img_array_in, np.ndarray):
        if img_array_in.ndim == 2:
            return tool(*args, **kwargs)  # Function call!

    for i, img in enumerate(img_array_in):
        args, kwargs = _replace_img_arg(img, *args, **kwargs)
        img_array_in[i] = tool(*args, **kwargs)  # Function call!

    return img_array_in


@decorator
def multiple_imgs_as_ndarray(tool, *args, **kwargs):
    """
    Images are processed as follows, esp. for temporal filters:
    .. array-like (input) --> nd-array ---> [`tool`] --> array_like (output)

    """
    img_array_in = _get_img_arg(*args, **kwargs)

    if isinstance(img_array_in, np.ndarray) and img_array_in.ndim == 3:
        return tool(*args, **kwargs)  # Function call!

    img_ndarray_in, ImgArrayLike = _get_array_like_type(img_array_in, True)
    args, kwargs = _replace_img_arg(img_ndarray_in, *args, **kwargs)

    img_array_out = tool(*args, **kwargs)  # Function call!
    return ImgArrayLike(img_array_out)



"""Topology for image2image preprocessing (:mod:`fluidimage.topologies.image2image`)
====================================================================================

.. autoclass:: TopologyImage2Image
   :members:
   :private-members:

"""

import json
import sys
from pathlib import Path

from fluiddyn.io.image import imsave
from fluidimage import ParamContainer, SerieOfArraysFromFiles
from fluidimage.preproc.image2image import (
    complete_im2im_params_with_default,
    init_im2im_function,
)
from fluidimage.topologies import prepare_path_dir_result
from fluidimage.util import imread, logger

from .base import TopologyBase


class TopologyImage2Image(TopologyBase):
    """Topology for images processing with a user-defined function

    The most useful methods for the user (in particular :func:`compute`) are
    defined in the base class :class:`fluidimage.topologies.base.TopologyBase`.

    Parameters
    ----------

    params : None

      A ParamContainer (created with the class method
      :func:`create_default_params`) containing the parameters for the
      computation.

    logging_level : str, {'warning', 'info', 'debug', ...}

      Logging level.

    nb_max_workers : None, int

      Maximum numbers of "workers". If None, a number is computed from the
      number of cores detected. If there are memory errors, you can try to
      decrease the number of workers.

    """

    @classmethod
    def create_default_params(cls):
        """Class method returning the default parameters.

        Typical usage::

           params = TopologyImage2Image.create_default_params()
           # modify parameters here
           ...

           topo = TopologyImage2Image(params)

        """
        params = ParamContainer(tag="params")
        complete_im2im_params_with_default(params)

        params._set_child("images", attribs={"path": "", "str_slice": None})

        params.images._set_doc(
            """
Parameters indicating the input image set.

path : str, {''}

    String indicating the input images (can be a full path towards an image
    file or a string given to `glob`).

str_slice : None

    String indicating as a Python slicing how to select images from the serie of
    images on the disk. If None, no selection so all images are going to be
    processed.

"""
        )

        params._set_child(
            "saving", attribs={"path": None, "how": "ask", "postfix": "pre"}
        )

        params.saving._set_doc(
            """Saving of the results.

path : None or str

    Path of the directory where the data will be saved. If None, the path is
    obtained from the input path and the parameter `postfix`.

how : str {'ask'}

    'ask', 'new_dir', 'complete' or 'recompute'.

postfix : str

    Postfix from which the output file is computed.
"""
        )

        params._set_internal_attr(
            "_value_text",
            json.dumps(
                {
                    "program": "fluidimage",
                    "module": "fluidimage.topologies.image2image",
                    "class": "TopologyImage2Image",
                }
            ),
        )

        return params

    def __init__(self, params, logging_level="info", nb_max_workers=None):

        self.params = params

        if params.im2im is None:
            raise ValueError("params.im2im has to be set.")

        self.serie = SerieOfArraysFromFiles(
            params.images.path, params.images.str_slice
        )

        path_dir = self.serie.path_dir
        path_dir_result, self.how_saving = prepare_path_dir_result(
            path_dir, params.saving.path, params.saving.postfix, params.saving.how
        )

        self.path_dir_result = path_dir_result
        self.path_dir_src = Path(path_dir)

        super().__init__(
            path_dir_result=path_dir_result,
            logging_level=logging_level,
            nb_max_workers=nb_max_workers,
        )

        self.queue_paths = self.add_queue("paths")
        self.queue_arrays = self.add_queue("arrays")
        self.queue_results = self.add_queue("results")

        self.add_work(
            "fill_path",
            self.fill_queue_paths,
            output_queue=self.queue_paths,
            kind="one shot",
        )

        self.add_work(
            "read_array",
            self.imread,
            input_queue=self.queue_paths,
            output_queue=self.queue_arrays,
            kind="io",
        )

        im2im_func = self.init_im2im(params)

        self.add_work(
            "im2im",
            im2im_func,
            input_queue=self.queue_arrays,
            output_queue=self.queue_results,
        )

        self.add_work(
            "save", self.save_image, input_queue=self.queue_results, kind="io"
        )

    def init_im2im(self, params_im2im):
        """Initialize the im2im function"""
        self.im2im_obj, self.im2im_func = init_im2im_function(
            im2im=params_im2im.im2im, args_init=params_im2im.args_init
        )
        return self.im2im_func

    def imread(self, path):
        """Read an image"""
        array = imread(path)
        return (array, path)

    def save_image(self, tuple_image_path):
        """Save an image"""
        image, path = tuple_image_path
        name_file = Path(path).name
        path_out = self.path_dir_result / name_file
        imsave(path_out, image)

    def fill_queue_paths(self, input_queue, output_queue):
        """Fill the first queue (paths)"""
        assert input_queue is None

        serie = self.serie
        if not serie:
            logger.warning("add 0 image. No image to process.")
            return

        names = serie.get_name_arrays()

        for name in names:
            path_im_output = self.path_dir_result / name
            path_im_input = str(self.path_dir_src / name)
            if self.how_saving == "complete":
                if not path_im_output.exists():
                    output_queue[name] = path_im_input
            else:
                output_queue[name] = path_im_input

        if not names:
            if self.how_saving == "complete":
                logger.warning(
                    'topology in mode "complete" and work already done.'
                )
            else:
                logger.warning("Nothing to do")
            return

        nb_names = len(names)

        logger.info(f"Add {nb_names} images to compute.")
        logger.info(f"First files to process: {names[:4]}")

        logger.debug(f"All files: {names}")


if "sphinx" in sys.modules:
    params = TopologyImage2Image.create_default_params()

    __doc__ += params._get_formatted_docs()



import unittest
from pathlib import Path
from shutil import rmtree

import skimage

from fluidimage import get_path_image_samples
from fluidimage.topologies.surface_tracking import TopologySurfaceTracking


def rescale_intensity(tuple_image_path):
    """
    Rescale image intensities, between the specified minima and maxima,
    by using a multiplicative factor.

    ----------
    minima, maxima : float
        Sets the range to which current intensities have to be rescaled.

    """
    img, path = tuple_image_path
    # the processing can be adjusted depending on the value of the path.
    print("process file:\n" + path)
    minima = 0
    maxima = 4095
    out_range = (minima, maxima)
    img_out = skimage.exposure.rescale_intensity(img, out_range=out_range)
    return img_out, path


class TestSurfaceTracking(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path_src = get_path_image_samples() / "SurfTracking/Images"
        cls.postfix = "test_surftracking_new"
        cls.path_out = Path(str(cls.path_src) + "." + cls.postfix)

    @classmethod
    def tearDownClass(cls):
        # path_out = Path(str(cls.path_src) + "." + cls.postfix)
        if cls.path_out.exists():
            rmtree(cls.path_out)

    def test_surftrack(self):
        params = TopologySurfaceTracking.create_default_params()

        params.images.path = str(self.path_src)
        params.images.path_ref = str(self.path_src)
        params.images.str_slice = ":4:2"
        params.images.str_slice_ref = ":3"

        params.surface_tracking.xmin = 200
        params.surface_tracking.xmax = 250
        params.surface_tracking.correct_pos = True

        params.preproc.im2im = rescale_intensity

        params.saving.how = "recompute"
        params.saving.path = str(self.path_out)
        params.saving.postfix = self.postfix

        topology = TopologySurfaceTracking(params, logging_level="info")
        topology.make_code_graphviz(topology.path_dir_result / "topo.dot")
        topology.compute(executor="exec_sequential", stop_if_error=True)
        # topology.compute(nb_max_workers=1)
        # topology.compute(sequential=True, stop_if_error=True)


if __name__ == "__main__":
    unittest.main()



"""
Multi executors async (:mod:`fluidimage.executors.multi_exec_async`)
====================================================================

.. autoclass:: MultiExecutorAsync
   :members:
   :private-members:

.. autoclass:: ExecutorAsyncForMulti
   :members:
   :private-members:

"""

import copy
import math
import os
import signal
import sys
from multiprocessing import Pipe, Process
from pathlib import Path
from time import time

from fluiddyn import time_as_str
from fluidimage.util import logger

from .base import ExecutorBase
from .exec_async_sequential import ExecutorAsyncSequential


class ExecutorAsyncForMulti(ExecutorAsyncSequential):
    """Slightly modified ExecutorAsync"""

    def __init__(
        self,
        topology,
        path_dir_result,
        log_path,
        sleep_time=0.01,
        logging_level="info",
        stop_if_error=False,
    ):
        if stop_if_error:
            raise NotImplementedError(
                "stop_if_error not implemented for ExecutorAsyncForMulti"
            )

        self._log_path = log_path
        super().__init__(
            topology,
            path_dir_result,
            nb_max_workers=1,
            nb_items_queue_max=8,
            sleep_time=sleep_time,
            logging_level=logging_level,
        )

    def _init_log_path(self):
        self.path_dir_exceptions = self._log_path.parent

    def _init_compute(self):
        self._init_compute_log()

    def _finalize_compute(self):
        self._reset_std_as_default()

        txt = self.topology.make_text_at_exit(time() - self.t_start)
        with open(self._log_file.name, "a") as file:
            file.write(txt)


class MultiExecutorAsync(ExecutorBase):
    """Manage the multi-executor mode

     This class is not the one whose really compute the topology. The topology is
     split and each slice is computed with an ExecutorAsync

    Parameters
    ----------

    nb_max_workers : None, int

      Limits the numbers of workers working in the same time.

    nb_items_queue_max : None, int

      Limits the numbers of items that can be in a output_queue.

    sleep_time : None, float

      defines the waiting time (from trio.sleep) of a function. Async functions
      await "trio.sleep(sleep_time)" when they have done a work on an item, and
      when there is nothing in there input_queue.

    """

    def __init__(
        self,
        topology,
        path_dir_result,
        nb_max_workers=None,
        nb_items_queue_max=None,
        sleep_time=0.01,
        logging_level="info",
        stop_if_error=False,
    ):
        if stop_if_error:
            raise NotImplementedError

        super().__init__(
            topology,
            path_dir_result,
            nb_max_workers,
            nb_items_queue_max,
            logging_level=logging_level,
        )

        self.sleep_time = sleep_time
        self.nb_processes = self.nb_max_workers
        self.processes = []

        # to avoid a pylint warning
        self.log_paths = None

    def _init_log_path(self):
        name = "_".join(("log", time_as_str(), str(os.getpid())))
        path_dir_log = self.path_dir_exceptions = self.path_dir_result / name
        path_dir_log.mkdir(exist_ok=True)
        self._log_path = path_dir_log / (name + ".txt")

    def compute(self):
        """Compute the topology.

        There are two ways to split self.topology work:

        - If first self.topology has "series" attribute (from seriesOfArray), it
          creates "self.nb_max_workers" topologies and changes "ind_start" and
          "ind_stop" of topology.series. The split considers series.ind_step.

        - Else, if the first work of the topology has an unique output_queue, it
          splits that queue in "self.nb_max_worker" slices and create as many
          topologies. On these last, the first work will be removed and the first
          queue will be filled with a partition of the first queue Then create as
          many Executer_await as topologies, give each topology to each executors,
          and call each Executor_await.compute in a process from multiprocessing.

        """
        self._init_compute()
        self.log_paths = []

        if sys.platform != "win32":

            def handler_signals(signal_number, stack):
                del stack
                print(
                    f"signal {signal_number} received: set _has_to_stop to True "
                    f"({type(self).__name__})."
                )
                self._has_to_stop = True
                for process in self.processes:
                    os.kill(process.pid, signal_number)

            signal.signal(12, handler_signals)

        if hasattr(self.topology, "series"):
            self.start_multiprocess_series()
        else:
            self.start_mutiprocess_first_queue()

        self._finalize_compute()

    def start_mutiprocess_first_queue(self):
        """Start the processes spitting the work with the first queue"""
        first_work = self.topology.works[0]
        if first_work.input_queue is not None:
            raise NotImplementedError
        if isinstance(first_work.output_queue, tuple):
            raise NotImplementedError

        # fill the first queue
        first_work.func_or_cls(
            input_queue=None, output_queue=first_work.output_queue
        )

        first_queue = copy.copy(first_work.output_queue)

        # split the first queue
        keys = list(first_queue.keys())

        nb_keys_per_process = max(1, int(len(keys) / self.nb_processes))

        keys_for_processes = []
        for iproc in range(self.nb_processes):
            istart = iproc * nb_keys_per_process
            keys_for_processes.append(keys[istart : istart + nb_keys_per_process])

        # change topology
        self.topology.first_queue = self.topology.works[0].output_queue
        topology = copy.copy(self.topology)
        topology.first_queue.clear()
        del topology.works[0]
        old_queue = topology.first_queue

        for ind_process, keys_proc in enumerate(keys_for_processes):
            topology_this_process = copy.copy(self.topology)
            new_queue = copy.copy(topology.first_queue)
            topology_this_process.first_queue = new_queue

            for iq, queue in enumerate(topology_this_process.queues):
                if queue is old_queue:
                    topology_this_process.queues[iq] = new_queue

            for work in topology_this_process.works:
                if work.output_queue is old_queue:
                    work.output_queue = new_queue

                if work.input_queue is old_queue:
                    work.input_queue = new_queue

                if isinstance(work.input_queue, (tuple, list)):
                    work.input_queue = list(work.input_queue)
                    for iq, queue in enumerate(work.input_queue):
                        if queue is old_queue:
                            work.input_queue[iq] = new_queue
                if isinstance(work.output_queue, (tuple, list)):
                    work.output_queue = list(work.output_queue)
                    for iq, queue in enumerate(work.output_queue):
                        if queue is old_queue:
                            work.output_queue[iq] = new_queue

            for key in keys_proc:
                new_queue[key] = first_queue[key]

            old_queue = new_queue

            self.launch_process(topology_this_process, ind_process)

        self.wait_for_all_processes()

    def start_multiprocess_series(self):
        """Start the processes spitting the work with the series object"""
        ind_stop_limit = self.topology.series.ind_stop
        # Defining split values
        ind_start = self.topology.series.ind_start
        nb_image_computed = math.floor(
            (self.topology.series.ind_stop - self.topology.series.ind_start)
            / self.topology.series.ind_step
        )
        remainder = nb_image_computed % self.nb_processes
        step_process = math.floor(nb_image_computed / self.nb_processes)
        # change topology
        for ind_process in range(self.nb_processes):
            new_topology = copy.copy(self.topology)
            new_topology.series.ind_start = ind_start
            add_rest = 0
            # To add forgotten images
            if remainder > 0:
                add_rest = 1
                remainder -= 1
            # defining ind_stop
            ind_stop = (
                self.topology.series.ind_start
                + step_process * self.topology.series.ind_step
                + add_rest * self.topology.series.ind_step
            )
            # To make sure images exist
            if ind_stop > ind_stop_limit:
                new_topology.series.ind_stop = ind_stop_limit
            else:
                new_topology.series.ind_stop = ind_stop
            ind_start = self.topology.series.ind_stop

            self.launch_process(new_topology, ind_process)

        self.wait_for_all_processes()

    def launch_process(self, topology, ind_process):
        """Launch one process"""

        def init_and_compute(topology_this_process, log_path, child_conn):
            """Create an executor and start it in a process"""
            executor = ExecutorAsyncForMulti(
                topology_this_process,
                self.path_dir_result,
                sleep_time=self.sleep_time,
                log_path=log_path,
                logging_level=self.logging_level,
            )
            executor.t_start = self.t_start
            executor.compute()

            # send the results
            if hasattr(topology_this_process, "results"):
                results = topology_this_process.results
            else:
                results = None

            child_conn.send(results)

        log_path = Path(
            str(self._log_path).split(".txt")[0] + f"_multi{ind_process:03}.txt"
        )

        self.log_paths.append(log_path)

        parent_conn, child_conn = Pipe()

        process = Process(
            target=init_and_compute, args=(topology, log_path, child_conn)
        )
        process.connection = parent_conn
        process.daemon = True
        process.start()
        self.processes.append(process)

    def wait_for_all_processes(self):
        """logging + wait for all processes to finish"""
        logger.info(
            f"logging files: {[log_path.name for log_path in self.log_paths]}"
        )

        # wait until end of all processes

        self.topology.results = results_all = []
        for process in self.processes:
            results = process.connection.recv()

            if results is not None:
                results_all.extend(results)

        for process in self.processes:
            process.join()



import pytest

from fluidimage import get_path_image_samples

try:
    import pyqtgraph
    from fluidimage.gui.pg_main import main, parse_args

    use_pyqtgraph = True
except ImportError:
    use_pyqtgraph = False

path_image_samples = get_path_image_samples()

# @pytest.mark.skipif(not use_pyqtgraph, reason="ImportError pyqtgraph")
# def test_main0(qtbot):
#     args = []
#     args.append(str(path_image_samples / "Jet" / "Images" / "c060a.png"))
#     args = parse_args(args)
#     pg = main(args, for_testing=True)
#     qtbot.addWidget(pg)


# @pytest.mark.skipif(not use_pyqtgraph, reason="ImportError pyqtgraph")
# def test_main(qtbot):
#     args = ["--slideshow"]
#     args.extend(
#         [
#             str(path)
#             for path in (path_image_samples / "Jet" / "Images").glob(
#                 "*a.png"
#             )
#         ]
#     )
#     args = parse_args(args)
#     pg = main(args, for_testing=True)
#     qtbot.addWidget(pg)



import unittest

from .util import cprint, imread, is_memory_full, str_short


class TestUtil(unittest.TestCase):
    def test_util(self):

        try:
            imread("__oups__.png")
        except (NameError, FileNotFoundError):
            print("IOError", IOError)

        cprint("nice")
        is_memory_full()
        str_short("string")


if __name__ == "__main__":
    unittest.main()



"""Find subpixel peak
=====================

.. autoclass:: SubPix
   :members:
   :private-members:

"""

import numpy as np

from transonic import boost

from .errors import PIVError


@boost
def compute_subpix_2d_gaussian2(correl: "float32[][]", ix: int, iy: int):
    correl_crop = correl[iy - 1 : iy + 2, ix - 1 : ix + 2]
    # hoops, pythran crashes because of this line
    # correl_crop[correl_crop < 0] = 1e-6

    # we write it like this to please pythran
    tmp = np.where(correl_crop < 0)
    for i0, i1 in zip(tmp[0], tmp[1]):
        correl_crop[i0, i1] = 1e-6

    c10 = 0
    c01 = 0
    c11 = 0
    c20 = 0
    c02 = 0
    for i in range(3):
        for j in range(3):
            c10 += (i - 1) * np.log(correl_crop[j, i])
            c01 += (j - 1) * np.log(correl_crop[j, i])
            c11 += (i - 1) * (j - 1) * np.log(correl_crop[j, i])
            c20 += (3 * (i - 1) ** 2 - 2) * np.log(correl_crop[j, i])
            c02 += (3 * (j - 1) ** 2 - 2) * np.log(correl_crop[j, i])
            c00 = (5 - 3 * (i - 1) ** 2 - 3 * (j - 1) ** 2) * np.log(
                correl_crop[j, i]
            )

    c00, c10, c01, c11, c20, c02 = (
        c00 / 9,
        c10 / 6,
        c01 / 6,
        c11 / 4,
        c20 / 6,
        c02 / 6,
    )
    deplx = (c11 * c01 - 2 * c10 * c02) / (4 * c20 * c02 - c11**2)
    deply = (c11 * c10 - 2 * c01 * c20) / (4 * c20 * c02 - c11**2)
    return deplx, deply, correl_crop


class SubPix:
    """Subpixel finder

    .. todo::

       - evaluate subpix methods.

    """

    methods = ["2d_gaussian", "2d_gaussian2", "centroid", "no_subpix"]

    def __init__(self, method="centroid", nsubpix=None):
        if method == "2d_gaussian2" and nsubpix is not None:
            raise ValueError(
                "Subpixel method '2d_gaussian2' doesn't require nsubpix. "
                "In this case, nsubpix has to be equal to None."
            )

        self.prepare_subpix(method, nsubpix)

    def prepare_subpix(self, method, nsubpix):
        self.method = method
        if nsubpix is None:
            nsubpix = 1
        self.n = nsubpix
        xs = ys = np.arange(-nsubpix, nsubpix + 1, dtype=float)
        X, Y = np.meshgrid(xs, ys)

        # init for centroid method
        self.X_centroid = X
        self.Y_centroid = Y

        # init for 2d_gaussian method
        nx, ny = X.shape
        X = X.ravel()
        Y = Y.ravel()
        M = np.reshape(
            np.concatenate((X**2, Y**2, X, Y, np.ones(nx * ny))), (5, nx * ny)
        ).T
        self.Minv_subpix = np.linalg.pinv(M)

    def compute_subpix(self, correl, ix, iy, method=None, nsubpix=None):
        """Find peak

        Parameters
        ----------

        correl: numpy.ndarray

          Normalized correlation

        ix: integer

        iy: integer

        method: str {'centroid', '2d_gaussian'}

        Notes
        -----

        The two methods...

        using linalg.solve (buggy?)

        """
        if method is None:
            method = self.method
        if nsubpix is None:
            nsubpix = self.n

        if method != self.method or nsubpix != self.n:
            if method is None:
                method = self.method
            if nsubpix is None:
                nsubpix = self.n
            self.prepare_subpix(method, nsubpix)

        if method not in self.methods:
            raise ValueError(f"method has to be in {self.methods}")

        ny, nx = correl.shape

        if (
            iy - nsubpix < 0
            or iy + nsubpix + 1 > ny
            or ix - nsubpix < 0
            or ix + nsubpix + 1 > nx
        ):
            raise PIVError(
                explanation="close boundary", result_compute_subpix=(iy, ix)
            )

        if method == "2d_gaussian":

            correl_crop = correl[
                iy - nsubpix : iy + nsubpix + 1, ix - nsubpix : ix + nsubpix + 1
            ]
            ny, nx = correl_crop.shape

            assert nx == ny == 2 * nsubpix + 1

            correl_map = correl_crop.ravel()
            correl_map[correl_map <= 0.0] = 1e-6

            coef = np.dot(self.Minv_subpix, np.log(correl_map))
            if coef[0] > 0 or coef[1] > 0:
                return self.compute_subpix(
                    correl, ix, iy, method="centroid", nsubpix=nsubpix
                )

            sigmax = 1 / np.sqrt(-2 * coef[0])
            sigmay = 1 / np.sqrt(-2 * coef[1])
            deplx = coef[2] * sigmax**2
            deply = coef[3] * sigmay**2

            if np.isnan(deplx) or np.isnan(deply):
                return self.compute_subpix(
                    correl, ix, iy, method="centroid", nsubpix=nsubpix
                )

        if method == "2d_gaussian2":
            deplx, deply, correl_crop = compute_subpix_2d_gaussian2(
                correl, int(ix), int(iy)
            )

        elif method == "centroid":

            correl_crop = correl[
                iy - nsubpix : iy + nsubpix + 1, ix - nsubpix : ix + nsubpix + 1
            ]
            ny, nx = correl_crop.shape

            sum_correl = np.sum(correl_crop)

            deplx = np.sum(self.X_centroid * correl_crop) / sum_correl
            deply = np.sum(self.Y_centroid * correl_crop) / sum_correl

        elif method == "no_subpix":
            deplx = deply = 0.0

        if deplx**2 + deply**2 > 2 * (0.5 + nsubpix) ** 2:
            if method == "2d_gaussian2":
                return self.compute_subpix(
                    correl, ix, iy, method="centroid", nsubpix=nsubpix
                )
            print(
                "Wrong subpix for one vector:"
                " deplx**2 + deply**2 > (0.5+nsubpix)**2\n"
                "method: " + method + f"\ndeplx, deply = ({deplx}, {deply})\n"
                f"correl_subpix =\n{correl_crop}"
            )
            raise PIVError(
                explanation="wrong subpix", result_compute_subpix=(iy, ix)
            )

        return deplx + ix, deply + iy



"""PIV objects (:mod:`fluidimage.data_objects.piv`)
===================================================

.. autoclass:: ArrayCouple
   :members:
   :private-members:

.. autoclass:: HeavyPIVResults
   :members:
   :private-members:

.. autoclass:: MultipassPIVResults
   :members:
   :private-members:

.. autoclass:: LightPIVResults
   :members:
   :private-members:

"""

import os

import h5netcdf
import h5py
import numpy as np

from fluidimage import ParamContainer, __version__ as fluidimage_version, imread
from fluidimage._version import hg_rev
from .display_piv import DisplayPIV


def get_str_index(serie, i, index):
    if serie._from_movies and i == serie.nb_indices - 1:
        return str(index)

    if serie._index_types[i] == "digit":
        code_format = "{:0" + str(serie._index_lens[i]) + "d}"
        str_index = code_format.format(index)
    elif serie._index_types[i] == "alpha":
        if index > 25:
            raise ValueError('"alpha" index larger than 25.')

        str_index = chr(ord("a") + index)
    else:
        raise Exception('The type should be "digit" or "alpha".')

    return str_index


def get_name_piv(serie, prefix="piv"):
    index_slices = serie._index_slices
    str_indices = ""
    for i, inds in enumerate(index_slices):
        index = inds[0]
        str_index = get_str_index(serie, i, index)
        if len(inds) > 1:
            str_index += "-" + get_str_index(serie, i, inds[1] - 1)

        if i > 1:
            str_indices += serie._index_separators[i - 1]
        str_indices += str_index

    name = prefix + "_" + str_indices + ".h5"
    return name


def get_name_bos(name, serie):
    name = name[len(serie.base_name) :]
    if serie.extension_file is not None:
        name = name[: -len(serie.extension_file) - 1]
    return "bos" + name + ".h5"


class DataObject:
    pass


def get_slices_from_strcrop(strcrop):
    return tuple(
        slice(*(int(i) if i else None for i in part.strip().split(":")))
        for part in strcrop.split(",")
    )


class ArrayCouple(DataObject):
    """Couple of arrays (images)."""

    def __init__(
        self,
        names=None,
        arrays=None,
        serie=None,
        str_path=None,
        hdf5_parent=None,
        params_mask=None,
    ):

        self.params_mask = params_mask

        if str_path is not None:
            self._load(path=str_path)
            return

        if hdf5_parent is not None:
            self._load(hdf5_object=hdf5_parent["couple"])
            return

        if serie is not None:
            names = serie.get_name_arrays()
            if len(names) != 2:
                raise ValueError("serie has to contain 2 arrays.")

            paths = serie.get_path_arrays()

            if not serie.check_all_files_exist():
                raise ValueError(
                    "At least one file of this serie does not exists. \n"
                    + str(paths)
                )

            self.paths = tuple(os.path.abspath(p) for p in paths)

            if arrays is None:
                arrays = self.read_images()

        self.names = tuple(names)
        self.name = "-".join(self.names)
        self.arrays = self._mask_arrays(arrays)
        self.serie = serie

    def _read_image(self, index):
        arr = imread(self.paths[index])
        return arr

    def read_images(self):
        return tuple(self._read_image(i) for i in [0, 1])

    def apply_mask(self, params_mask):
        if self.params_mask is not None and params_mask is None:
            raise NotImplementedError

        if self.params_mask is not None and params_mask == self.params_mask:
            return

        if self.params_mask is not None:
            raise NotImplementedError

        self.params_mask = params_mask
        self.arrays = self._mask_arrays(self.arrays)

    def _mask_array(self, array):
        if self.params_mask is None:
            return array

        if self.params_mask.strcrop is not None:
            indices = get_slices_from_strcrop(self.params_mask.strcrop)
            array = array[indices]
        return array

    def _mask_arrays(self, arrays):
        return tuple(self._mask_array(arr) for arr in arrays)

    def get_arrays(self):
        if not hasattr(self, "arrays"):
            self.arrays = self._mask_arrays(self.read_images())

        return self.arrays

    def save(self, path=None, hdf5_parent=None):
        if path is not None:
            raise NotImplementedError

        if not isinstance(hdf5_parent, (h5py.File,)):
            raise NotImplementedError

        hdf5_parent.create_group("couple")
        group = hdf5_parent["couple"]

        assert isinstance(self.names, tuple)
        assert isinstance(self.paths, tuple)

        group.attrs["names"] = repr(self.names).encode()
        group.attrs["paths"] = repr(self.paths).encode()

        if not hasattr(self, "arrays"):
            arr0 = self._mask_array(self._read_image(0))
        else:
            arr0 = self.arrays[0]

        group.create_dataset("shape_images", data=arr0.shape)

    def _load(self, path=None, hdf5_object=None):

        if path is not None:
            raise NotImplementedError

        names = hdf5_object.attrs["names"]
        paths = hdf5_object.attrs["paths"]

        if isinstance(names, bytes):
            names = names.decode()
            paths = paths.decode()

        self.names = eval(names)
        self.paths = eval(paths)

        assert isinstance(self.names, tuple)
        assert isinstance(self.paths, tuple)

        try:
            self.shape_images = hdf5_object["shape_images"][...]
        except KeyError:
            pass


class ArrayCoupleBOS(ArrayCouple):
    """Couple of arrays (images)."""

    def __init__(
        self,
        names=None,
        arrays=None,
        serie=None,
        paths=None,
        str_path=None,
        hdf5_parent=None,
        params_mask=None,
    ):

        self.params_mask = params_mask

        if str_path is not None:
            self._load(path=str_path)
            return

        if hdf5_parent is not None:
            self._load(hdf5_object=hdf5_parent["couple"])
            return

        if paths is not None:
            self.paths = tuple(os.path.abspath(p) for p in paths)

        self.serie = serie
        self.names = tuple(names)
        self.name = self.names[-1]
        self.arrays = self._mask_arrays(arrays)


class HeavyPIVResults(DataObject):
    """Heavy PIV results containing displacements and correlation."""

    _keys_to_be_saved = [
        "xs",
        "ys",
        "deltaxs",
        "deltays",
        "correls_max",
        "deltaxs_approx",
        "deltays_approx",
        "ixvecs_approx",
        "iyvecs_approx",
        "deltaxs_final",
        "deltays_final",
        "ixvecs_final",
        "iyvecs_final",
    ]

    _dict_to_be_saved = ["errors", "deltaxs_wrong", "deltays_wrong"]

    def __init__(
        self,
        deltaxs=None,
        deltays=None,
        xs=None,
        ys=None,
        errors=None,
        correls_max=None,
        correls=None,
        couple=None,
        params=None,
        str_path=None,
        hdf5_object=None,
        secondary_peaks=None,
        indices_no_displacement=None,
    ):

        if hdf5_object is not None:
            if couple is not None:
                self.couple = couple

            if params is not None:
                self.params = params

            self._load_from_hdf5_object(hdf5_object)
            return

        if str_path is not None:
            self._load(str_path)
            return

        self.deltaxs = deltaxs
        self.deltays = deltays
        self.ys = ys
        self.xs = xs
        self.errors = errors
        self.correls_max = correls_max
        self.correls = correls
        self.couple = couple
        self.params = params
        self.secondary_peaks = secondary_peaks
        self.indices_no_displacement = indices_no_displacement

    def get_images(self):
        return self.couple.read_images()

    def display(
        self,
        show_interp=False,
        scale=0.2,
        show_error=True,
        pourcent_histo=99,
        hist=False,
        show_correl=True,
    ):
        try:
            im0, im1 = self.get_images()
        except IOError:
            im0, im1 = None, None
        return DisplayPIV(
            im0,
            im1,
            self,
            show_interp=show_interp,
            scale=scale,
            show_error=show_error,
            pourcent_histo=pourcent_histo,
            hist=hist,
            show_correl=show_correl,
        )

    def _get_name(self, kind):
        serie = self.couple.serie

        if kind is None:
            return get_name_piv(serie, prefix="piv")

        elif kind == "bos":
            return get_name_bos(self.couple.name, serie)

    def save(self, path=None, out_format=None, kind=None):

        name = self._get_name(kind)

        if path is not None:
            path_file = os.path.join(path, name)
        else:
            path_file = name

        if out_format == "uvmat":
            with h5netcdf.File(path_file, "w") as file:
                self._save_as_uvmat(file)
        else:
            with h5py.File(path_file, "w") as file:
                file.attrs["class_name"] = "HeavyPIVResults"
                file.attrs["module_name"] = "fluidimage.data_objects.piv"
                self._save_in_hdf5_object(file)

        return path_file

    def _save_in_hdf5_object(self, file, tag="piv0"):

        if "class_name" not in file.attrs.keys():
            file.attrs["class_name"] = "HeavyPIVResults"
            file.attrs["module_name"] = "fluidimage.data_objects.piv"

        if "params" not in file.keys():
            self.params._save_as_hdf5(hdf5_parent=file)

        if "couple" not in file.keys():
            self.couple.save(hdf5_parent=file)

        g_piv = file.create_group(tag)
        g_piv.attrs["class_name"] = "HeavyPIVResults"
        g_piv.attrs["module_name"] = "fluidimage.data_objects.piv"

        for k in self._keys_to_be_saved:
            if k in self.__dict__ and self.__dict__[k] is not None:
                g_piv.create_dataset(k, data=self.__dict__[k])

        for name_dict in self._dict_to_be_saved:
            try:
                d = self.__dict__[name_dict]
                if d is None:
                    raise KeyError
            except KeyError:
                pass
            else:
                g = g_piv.create_group(name_dict)
                keys = list(d.keys())
                values = list(d.values())
                try:
                    for i, k in enumerate(keys):
                        keys[i] = k.encode()
                except AttributeError:
                    pass
                try:
                    for i, k in enumerate(values):
                        values[i] = k.encode()
                except AttributeError:
                    pass

                g.create_dataset("keys", data=keys)
                g.create_dataset("values", data=values)

        if "deltaxs_tps" in self.__dict__:
            g = g_piv.create_group("deltaxs_tps")
            for i, arr in enumerate(self.deltaxs_tps):
                g.create_dataset(f"subdom{i}", data=arr)

            g = g_piv.create_group("deltays_tps")
            for i, arr in enumerate(self.deltays_tps):
                g.create_dataset(f"subdom{i}", data=arr)

    def _load(self, path):

        self.file_name = os.path.basename(path)
        with h5py.File(path, "r") as file:
            self._load_from_hdf5_object(file["piv0"])

    def _load_from_hdf5_object(self, g_piv):

        file = g_piv.parent

        if not hasattr(self, "params"):
            self.params = ParamContainer(hdf5_object=file["params"])
            try:
                params_mask = self.params.mask
            except AttributeError:
                params_mask = None

        if not hasattr(self, "couple"):
            self.couple = ArrayCouple(hdf5_parent=file, params_mask=params_mask)

        for k in self._keys_to_be_saved:
            if k in g_piv:
                dataset = g_piv[k]
                self.__dict__[k] = dataset[:]

        for name_dict in self._dict_to_be_saved:
            try:
                g = g_piv[name_dict]
                keys = g["keys"]
                values = g["values"]
                dictionary = {}
                for key, value in zip(keys, values):
                    try:
                        key = k.decode()
                    except AttributeError:
                        pass
                    try:
                        value = value.decode()
                    except AttributeError:
                        pass
                    dictionary[key] = value
                self.__dict__[name_dict] = dictionary

            except KeyError:
                pass

        if "deltaxs_tps" in g_piv.keys():
            g = g_piv["deltaxs_tps"]
            self.deltaxs_tps = []
            for arr in g.keys():
                self.deltaxs_tps.append(g[arr][...])
            g = g_piv["deltays_tps"]
            self.deltays_tps = []
            for arr in g.keys():
                self.deltays_tps.append(g[arr][...])

    def get_grid_pixel(self, index_pass):
        """Recompute 1d arrays containing the approximate positions of the vectors

        Useful to compute a grid on which we can interpolate the displacement fields.

        Parameters
        ----------

        index_pass: int

          Index of the pass

        Returns
        -------

        xs1d: np.ndarray

          Indices (2nd, direction "x") of the pixel in the image

        ys1d: np.ndarray

          Indices (1st, direction "y") of the pixel in the image

        """
        from ..postproc.piv import get_grid_pixel

        return get_grid_pixel(self.params, self.couple.shape_images, index_pass)


class MultipassPIVResults(DataObject):
    """Result of a multipass PIV computation."""

    def __init__(self, str_path=None):
        self.passes = []

        if str_path is not None:
            self._load(str_path)

    def display(
        self,
        i=-1,
        show_interp=False,
        scale=0.2,
        show_error=True,
        pourcent_histo=99,
        hist=False,
        show_correl=True,
    ):
        r = self.passes[i]
        return r.display(
            show_interp=show_interp,
            scale=scale,
            show_error=show_error,
            pourcent_histo=pourcent_histo,
            hist=hist,
            show_correl=show_correl,
        )

    def __getitem__(self, key):
        return self.passes[key]

    def append(self, results):
        i = len(self.passes)
        self.passes.append(results)
        self.__dict__[f"piv{i}"] = results

    def _get_name(self, kind):

        if hasattr(self, "file_name"):
            return self.file_name

        r = self.passes[0]
        return r._get_name(kind)

    def save(self, path=None, out_format=None, kind=None):

        name = self._get_name(kind)

        if path is not None:
            root, ext = os.path.splitext(path)
            if ext in [".h5", ".nc"]:
                path_file = path
            else:
                path_file = os.path.join(path, name)
        else:
            path_file = name

        if out_format == "uvmat":
            with h5netcdf.File(path_file, "w") as file:
                self._save_as_uvmat(file)
        else:
            with h5py.File(path_file, "w") as file:
                file.attrs["class_name"] = "MultipassPIVResults"
                file.attrs["module_name"] = "fluidimage.data_objects.piv"

                file.attrs["nb_passes"] = len(self.passes)

                file.attrs["fluidimage_version"] = fluidimage_version
                file.attrs["fluidimage_hg_rev"] = hg_rev

                for i, r in enumerate(self.passes):
                    r._save_in_hdf5_object(file, tag=f"piv{i}")

        return path_file

    def _load(self, path):

        self.file_name = os.path.basename(path)
        with h5py.File(path, "r") as file:
            self.params = ParamContainer(hdf5_object=file["params"])

            try:
                params_mask = self.params.mask
            except AttributeError:
                params_mask = None

            self.couple = ArrayCouple(hdf5_parent=file, params_mask=params_mask)

            nb_passes = file.attrs["nb_passes"]

            for ip in range(nb_passes):
                g = file[f"piv{ip}"]
                self.append(
                    HeavyPIVResults(
                        hdf5_object=g, couple=self.couple, params=self.params
                    )
                )

    def _save_as_uvmat(self, file):

        file.dimensions = {"nb_coord": 2, "nb_bounds": 2}

        for i, ir in enumerate(self.passes):

            iuvmat = i + 1
            str_i = str(iuvmat)
            str_nb_vec = "nb_vec_" + str_i
            file.dimensions[str_nb_vec] = ir.xs.size

            tmp = np.zeros(ir.deltaxs.shape).astype("float32")
            inds = np.where(~np.isnan(ir.deltaxs))

            file.create_variable(f"Civ{iuvmat}_X", (str_nb_vec,), data=ir.xs)
            file.create_variable(f"Civ{iuvmat}_Y", (str_nb_vec,), data=ir.ys)
            file.create_variable(
                f"Civ{iuvmat}_U", (str_nb_vec,), data=np.nan_to_num(ir.deltaxs)
            )
            file.create_variable(
                f"Civ{iuvmat}_V", (str_nb_vec,), data=np.nan_to_num(ir.deltays)
            )

            if ir.params.multipass.use_tps:
                str_nb_subdom = f"nb_subdomain_{iuvmat}"
                try:
                    file.dimensions[str_nb_subdom] = np.shape(ir.deltaxs_tps)[0]
                    file.dimensions[f"nb_tps{iuvmat}"] = np.shape(ir.deltaxs_tps)[
                        1
                    ]
                    tmp[inds] = ir.deltaxs_smooth
                    file.create_variable(
                        f"Civ{iuvmat}_U_smooth", (str_nb_vec,), data=tmp
                    )
                    tmp[inds] = ir.deltays_smooth
                    file.create_variable(
                        f"Civ{iuvmat}_V_smooth", (str_nb_vec,), data=tmp
                    )
                    file.create_variable(
                        f"Civ{iuvmat}_U_tps",
                        (str_nb_subdom, f"nb_vec_tps_{iuvmat}"),
                        data=ir.deltaxs_tps,
                    )
                    file.create_variable(
                        f"Civ{iuvmat}_V_tps",
                        (str_nb_subdom, f"nb_vec_tps_{iuvmat}"),
                        data=ir.deltays_tps,
                    )
                    tmp = [None] * file.dimensions[str_nb_subdom]
                    for j in range(file.dimensions[str_nb_subdom]):
                        tmp[j] = np.shape(ir.deltaxs_tps[j])[0]
                    file.create_variable(
                        f"Civ{iuvmat}_NbCentres",
                        (f"nb_subdomain_{iuvmat}",),
                        data=tmp,
                    )
                except:
                    print(f"no tps field at passe n {iuvmat}")

            file.create_variable(
                f"Civ{iuvmat}_C", (str_nb_vec,), data=ir.correls_max
            )
            tmp = np.zeros(ir.deltaxs.shape).astype("float32")
            indsnan = np.where(np.isnan(ir.deltaxs))
            tmp[indsnan] = 1
            file.create_variable(f"Civ{iuvmat}_FF", (str_nb_vec,), data=tmp)

            # mettre bonne valeur de F correspondant a self.piv0.error...
            file.create_variable(f"Civ{iuvmat}_F", (str_nb_vec,), data=tmp)

    # ADD
    # file.create_variable('Civ1_Coord_tps',
    #                   ('nb_subdomain_1', 'nb_coord', 'nb_tps_1'),
    #                   data=???)

    # ADD attributes

    def make_light_result(self, ind_pass=-1):
        piv = self.passes[ind_pass]
        if ind_pass == -1:
            deltaxs = piv.deltaxs_final
            deltays = piv.deltays_final
            ixvec = piv.ixvecs_final
            iyvec = piv.iyvecs_final

        else:
            deltaxs = piv.deltaxs_approx
            deltays = piv.deltays_approx
            ixvec = piv.ixvecs_approx
            iyvec = piv.ixvecs_approx

        if hasattr(self, "file_name"):
            file_name = self.file_name
        else:
            file_name = None
        return LightPIVResults(
            deltaxs,
            deltays,
            ixvec,
            iyvec,
            couple=piv.couple,
            params=piv.params,
            file_name=file_name,
        )

    def get_grid_pixel(self, index_pass=-1):
        """Recompute 1d arrays containing the approximate positions of the vectors

        Useful to compute a grid on which we can interpolate the displacement fields.

        Parameters
        ----------

        index_pass: int

          Index of the pass

        Returns
        -------

        xs1d: np.ndarray

          Indices (2nd, direction "x") of the pixel in the image

        ys1d: np.ndarray

          Indices (1st, direction "y") of the pixel in the image

        """
        piv = self.passes[index_pass]
        return piv.get_grid_pixel(index_pass)


class LightPIVResults(DataObject):
    _keys_to_be_saved = [
        "ixvecs_final",
        "iyvecs_final",
        "deltaxs_final",
        "deltays_final",
    ]

    def __init__(
        self,
        deltaxs_approx=None,
        deltays_approx=None,
        ixvecs_grid=None,
        iyvecs_grid=None,
        correls_max=None,
        couple=None,
        params=None,
        str_path=None,
        hdf5_object=None,
        file_name=None,
    ):

        if file_name is not None:
            self.file_name = file_name
        if hdf5_object is not None:
            if couple is not None:
                self.couple = couple

            if params is not None:
                self.params = params

            self._load_from_hdf5_object(hdf5_object)
            return

        if str_path is not None:
            self._load(str_path)
            return

        self.deltaxs_final = deltaxs_approx
        self.deltays_final = deltays_approx
        self.couple = couple
        self.params = params
        self.ixvecs_final = ixvecs_grid
        self.iyvecs_final = iyvecs_grid

    def _get_name(self, kind):
        if hasattr(self, "file_name"):
            return self.file_name[:-3] + "_light.h5"

        serie = self.couple.serie

        str_ind0 = serie._compute_strindices_from_indices(
            *[inds[0] for inds in serie.get_index_slices()]
        )

        str_ind1 = serie._compute_strindices_from_indices(
            *[inds[1] - 1 for inds in serie.get_index_slices()]
        )

        name = "piv_" + serie.base_name + str_ind0 + "-" + str_ind1 + "_light.h5"
        return name

    def save(self, path=None, out_format=None, kind=None):
        name = self._get_name(kind)

        if path is not None:
            path_file = os.path.join(path, name)
        else:
            path_file = name

        with h5py.File(path_file, "w") as file:
            file.attrs["class_name"] = "LightPIVResults"
            file.attrs["module_name"] = "fluidimage.data_objects.piv"

            self._save_in_hdf5_object(file, tag="piv")

        return self

    def _save_in_hdf5_object(self, file, tag="piv"):

        if "class_name" not in file.attrs.keys():
            file.attrs["class_name"] = "LightPIVResults"
            file.attrs["module_name"] = "fluidimage.data_objects.piv"
        if "params" not in file.keys():
            self.params._save_as_hdf5(hdf5_parent=file)
        if "couple" not in file.keys():
            self.couple.save(hdf5_parent=file)

        g_piv = file.create_group(tag)
        g_piv.attrs["class_name"] = "LightPIVResults"
        g_piv.attrs["module_name"] = "fluidimage.data_objects.piv"

        for k in self._keys_to_be_saved:
            if k in self.__dict__:
                g_piv.create_dataset(k, data=self.__dict__[k])

    def _load(self, path):
        with h5py.File(path, "r") as file:
            self.params = ParamContainer(hdf5_object=file["params"])
            try:
                params_mask = self.params.mask
            except AttributeError:
                params_mask = None
            self.couple = ArrayCouple(hdf5_parent=file, params_mask=params_mask)

        with h5py.File(path, "r") as file:
            keys = [(key) for key in file.keys() if "piv" in key]
            self._load_from_hdf5_object(file[max(keys)])

    def _load_from_hdf5_object(self, g_piv):

        file = g_piv.parent

        if not hasattr(self, "params"):
            self.params = ParamContainer(hdf5_object=file["params"])
            try:
                params_mask = self.params.mask
            except AttributeError:
                params_mask = None

        if not hasattr(self, "couple"):
            self.couple = ArrayCouple(hdf5_parent=file, params_mask=params_mask)

        for k in self._keys_to_be_saved:
            if k in g_piv:
                dataset = g_piv[k]
                self.__dict__[k] = dataset[:]



import unittest
from functools import partialmethod
from shutil import rmtree

from fluidimage import get_path_image_samples
from fluidimage.topologies import LogTopology
from fluidimage.topologies.example import TopologyExample

path_input = get_path_image_samples() / "Karman/Images"

executors = [
    "exec_sequential",
    "exec_async_sequential",
    "exec_async",
    "multi_exec_async",
    "exec_async_multi",
    "exec_async_servers",
    "exec_async_servers_threading",
]


def _test(self, executor=None):

    params = TopologyExample.create_default_params()
    params["path_input"] = path_input

    path_dir_result = path_input.parent / f"Images.{executor}"
    params["path_dir_result"] = path_dir_result

    self.topology = topology = TopologyExample(params, logging_level="debug")
    topology.compute(executor, nb_max_workers=2)

    if executor != "exec_async_servers_threading":
        # there is a logging problem with this class but we don't mind.
        log = LogTopology(path_dir_result)
        self.assertTrue(log.topology_name is not None)

    path_files = tuple(path_dir_result.glob("Karman*"))

    assert len(path_files) > 0, "No files saved"
    assert len(path_files) == 2, "Bad number of saved files"


class TestTopoExample(unittest.TestCase):
    def tearDown(self):
        rmtree(self.topology.path_dir_result, ignore_errors=True)


for executor in executors:
    setattr(
        TestTopoExample,
        "test_" + str(executor),
        partialmethod(_test, executor=executor),
    )


if __name__ == "__main__":
    unittest.main()



"""Topology for optical flow (:mod:`fluidimage.topologies.optical_flow`)
========================================================================

.. autoclass:: TopologyOpticalFlow
   :members:
   :private-members:


"""

import sys

from fluidimage.works.optical_flow import WorkOpticalFlow

from .piv import TopologyPIV


class TopologyOpticalFlow(TopologyPIV):
    """Optical flow topology (Lukas Kanade method)"""

    WorkVelocimetry = WorkOpticalFlow

    @classmethod
    def create_default_params(cls):
        """Class method returning the default parameters.

        Typical usage::

          params = TopologyOpticalFlow.create_default_params()
          # modify parameters here
          ...

          topo = TopologyOpticalFlow(params)

        """

        params = super().create_default_params()
        params.saving.postfix = "optflow"
        return params


Topology = TopologyOpticalFlow

if "sphinx" in sys.modules:
    params = Topology.create_default_params()
    __doc__ += params._get_formatted_docs()


if __name__ == "__main__":
    from fluidimage import get_path_image_samples

    params = Topology.create_default_params()

    params.series.path = str(get_path_image_samples() / "Karman/Images")
    params.series.ind_start = 1
    params.series.ind_step = 2

    params.mask.strcrop = ":, 50:500"

    # params.preproc.im2im = "numpy.ones_like"

    # params.saving.how = 'complete'
    params.saving.postfix = "optical_flow_example"

    topo = Topology(params, logging_level="info")

    # topo.make_code_graphviz("tmp.dot")
    topo.compute()



import unittest
from shutil import rmtree

from fluiddyn.io import stdout_redirected
from fluidimage import SeriesOfArrays, get_path_image_samples
from fluidimage.data_objects.display_piv import DisplayPIV
from fluidimage.data_objects.piv import LightPIVResults, MultipassPIVResults
from fluidimage.works.piv import WorkPIV


class MyObj:
    pass


class TestPIV(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        path_images = get_path_image_samples() / "Oseen/Images"
        series = SeriesOfArrays(str(path_images / "Oseen*"), "i+1:i+3")
        cls.path_tmp = path_images.parent / "tmp_test_work_piv"

        if not cls.path_tmp.exists():
            cls.path_tmp.mkdir()

        cls.serie = series.get_serie_from_index(0)

    @classmethod
    def tearDownClass(cls):
        if cls.path_tmp.exists():
            rmtree(cls.path_tmp)

    def test_minimal_piv(self):

        params = WorkPIV.create_default_params()

        # for a very short computation
        params.piv0.shape_crop_im0 = 32
        params.piv0.grid.overlap = -3

        # still buggy
        # params.piv0.nb_peaks_to_search = 2

        params.multipass.number = 2

        params.fix.displacement_max = 2
        params.fix.threshold_diff_neighbour = 2

        piv = WorkPIV(params=params)

        with stdout_redirected():
            result = piv.calcul(self.serie)

        result.piv0.save(self.path_tmp)
        result.save(self.path_tmp)

        path_file = next(self.path_tmp.iterdir())
        MultipassPIVResults(path_file)

        result.save(self.path_tmp, "uvmat")

        light = result.make_light_result()

        light.save(self.path_tmp)
        path_file = next(self.path_tmp.glob("*light*"))

        LightPIVResults(str_path=str(path_file))

    def test_piv_list(self):

        params = WorkPIV.create_default_params()

        # for a very short computation
        params.piv0.shape_crop_im0 = [33, 44]
        params.piv0.grid.overlap = -3
        # params.piv0.nb_peaks_to_search = 2

        params.multipass.use_tps = False

        piv = WorkPIV(params=params)

        with stdout_redirected():
            result = piv.calcul(self.serie)

        piv0 = result.piv0
        im0, im1 = piv0.get_images()
        with stdout_redirected():
            DisplayPIV(im0, im1, piv0)
            d = DisplayPIV(im0, im1, piv0, show_interp=True, hist=True)

        d.switch()
        d.select_arrow([0], artist=d.q)
        d.select_arrow([0], artist=None)

        event = MyObj()

        event.key = "alt+h"
        event.inaxes = None
        d.onclick(event)

        event.inaxes = d.ax1

        event.key = "alt+s"
        d.onclick(event)
        event.key = "alt+left"
        d.onclick(event)
        event.key = "alt+right"
        d.onclick(event)

        event.artist = None
        d.onpick(event)

        event.artist = d.q
        event.ind = [0]
        d.onpick(event)


if __name__ == "__main__":
    unittest.main()



"""Topology base (:mod:`fluidimage.topologies.base`)
====================================================

.. autoclass:: Work
   :members:
   :private-members:

.. autoclass:: Queue
   :members:
   :private-members:

.. autoclass:: TopologyBase
   :members:
   :private-members:

"""

from collections import OrderedDict
from warnings import warn

from fluidimage.util import cstring, logger

from ..executors import ExecutorBase, executors


class Work:
    """Represent a work"""

    def __init__(
        self,
        name: str,
        func_or_cls,
        params_cls=None,
        input_queue=None,
        output_queue=None,
        kind: str = None,
    ):
        self._kwargs = dict(
            name=name,
            func_or_cls=func_or_cls,
            params_cls=params_cls,
            input_queue=input_queue,
            output_queue=output_queue,
            kind=kind,
        )
        # to avoid a pylint warning
        self.name = None

        self.__dict__.update(self._kwargs)
        self.name_no_space = self.name.replace(" ", "_")

    def __repr__(self):
        return super().__repr__() + f"\n{self._kwargs}"

    def check_exception(self, key, obj):
        """Check if `obj` is an exception"""
        if isinstance(obj, Exception):
            if self.output_queue is not None:
                self.output_queue[key] = obj
            else:
                logger.error(
                    cstring(
                        f"work {self.name_no_space} ({key}) "
                        "can not be done because of a previously "
                        "raised exception.",
                        color="FAIL",
                    )
                )
            return True
        return False


class Queue(OrderedDict):
    """Represent a queue"""

    def __init__(self, name, kind=None):
        self.name = name
        self.kind = kind
        super().__init__()

    def __repr__(self):
        return f'\nqueue "{self.name}": ' + super().__repr__()

    def __copy__(self):
        newone = type(self)(self.name, kind=self.kind)
        newone.__dict__.update(self.__dict__)

        for key, values in self.items():
            newone[key] = values

        return newone

    def pop_first_item(self):
        return self.popitem(last=False)


class TopologyBase:
    """Base class for topologies of processing.

    This class is meant to be subclassed, not instantiated directly.

    Parameters
    ----------

    path_dir_result : None, str

    logging_level : None,  {'warning', 'info', 'debug', ...}

    nb_max_workers : None, int

    """

    def __init__(
        self, path_dir_result=None, logging_level="info", nb_max_workers=None
    ):

        self.path_dir_result = path_dir_result
        self.logging_level = logging_level
        self.nb_max_workers = nb_max_workers

        self.queues = []
        self.works = []
        self.works_dict = {}
        self.executor = None

    def add_queue(self, name: str, kind: str = None):
        """Create a new queue."""
        queue = Queue(name=name, kind=kind)
        self.queues.append(queue)
        return queue

    def add_work(
        self,
        name: str,
        func_or_cls,
        params_cls=None,
        input_queue=None,
        output_queue=None,
        kind: str = None,
    ):
        """Create a new work relating queues."""
        if func_or_cls is None:
            warn(f'func_or_cls is None for work "{name}"')

        work = Work(
            name=name,
            input_queue=input_queue,
            func_or_cls=func_or_cls,
            params_cls=params_cls,
            output_queue=output_queue,
            kind=kind,
        )
        self.works.append(work)

        if name in self.works_dict:
            raise ValueError(f"The name {name} is already used.")
        self.works_dict[name] = work

    def compute(
        self,
        executor=None,
        nb_max_workers=None,
        sleep_time=0.01,
        sequential=False,
        stop_if_error=False,
    ):
        """Compute (run the works until all queues are empty).

        Parameters
        ----------

        executor : str or fluidimage.executors.base.ExecutorBase, optional

          If None, ``executor="multi_exec_async"``

        nb_max_workers : int, optional

        sleep_time : number, optional {0.01}

        sequential : bool, optional {False}

        stop_if_error : bool, optional {False}

        """

        if sequential:
            if executor is not None and executor != "exec_sequential":
                raise ValueError(
                    "Incompatible arguments sequential=True and "
                    f"executor={executor}"
                )
            executor = "exec_sequential"

        if executor is None:
            # fastest and safest executor for most cases
            executor = "multi_exec_async"

        if not isinstance(executor, ExecutorBase):
            if executor not in executors:
                raise NotImplementedError(f"executor {executor} does not exist")

            exec_class = executors[executor]
            self.executor = exec_class(
                self,
                path_dir_result=self.path_dir_result,
                nb_max_workers=nb_max_workers,
                sleep_time=sleep_time,
                logging_level=self.logging_level,
                stop_if_error=stop_if_error,
            )

        self.executor.compute()

    def make_text_at_exit(self, time_since_start):
        """Make a text printed before exit."""
        txt = f"Stop compute after t = {time_since_start:.2f} s"
        try:
            nb_results = len(self.results)
        except AttributeError:
            nb_results = None
        if nb_results is not None and nb_results > 0:
            txt += " ({} results, {:.2f} s/result).".format(
                nb_results, time_since_start / nb_results
            )
        else:
            txt += "."

        if hasattr(self, "path_dir_result"):
            txt += "\npath results:\n" + str(self.path_dir_result)

        return txt

    def print_at_exit(self, time_since_start):
        """Print information before exit."""
        print(self.make_text_at_exit(time_since_start))

    def make_code_graphviz(self, name_file="tmp.dot"):
        """Generate the graphviz / dot code.

        This method only generates a graphviz code. The graph can be visualized
        with for example::

          dot {name_file}.dot -Tpng -o {name_file}.png && eog {name_file}.png

        """
        name_file = str(name_file)

        if name_file.endswith(".dot"):
            name_file = name_file[:-4]

        code = "digraph {\nrankdir = LR\ncompound=true\n"
        # waiting queues
        code += '\nnode [shape="record"]\n'
        txt_queue = (
            '{name_quoted:40s} [label="<f0> {name}|'
            + "|".join([f"<f{i}>" for i in range(1, 5)])
            + '"]\n'
        )

        for queue in self.queues:
            name_quoted = f'"{queue.name}"'
            code += txt_queue.format(name=queue.name, name_quoted=name_quoted)

        # works and links
        code += '\nnode [shape="ellipse"]\n'

        txt_work = '{:40s} [label="{}",color = "{}"]\n'

        for work in self.works:
            name_work = work.name
            color = "Black"
            if work.kind is not None:
                if "io" in work.kind:
                    color = "Green"
            code += txt_work.format(f'"{name_work}"', name_work, color)

        code += "\n"

        str_link = (
            '{:40s} -> "{}" [arrowhead = "{}", style = "{}", color = "{}"]\n'
        )

        for work in self.works:
            name_work = work.name
            arrowhead = "normal"
            style = "dashed"
            color = "Black"
            if work.kind is not None:
                if "one shot" in work.kind:
                    style = "filled"
                if "global" in work.kind:
                    arrowhead = "odiamond"
                if "io" in work.kind:
                    color = "Green"
            if work.input_queue is not None:
                queues = work.input_queue
                if isinstance(queues, Queue):
                    queues = (queues,)
                for queue in queues:
                    code += str_link.format(
                        '"' + queue.name + '"', name_work, arrowhead, style, color
                    )
            if work.output_queue is not None:
                queues = work.output_queue
                if isinstance(queues, Queue):
                    queues = (queues,)
                for queue in queues:
                    code += str_link.format(
                        '"' + name_work + '"', queue.name, arrowhead, style, color
                    )

        # Legend
        code += "\n subgraph cluster_01 {"
        code += '\n node [height="0px", width="0px",shape=none,];'
        code += "\n edge [ minlen = 1,];"
        code += '\n label = "Legend";'
        code += '\n key [label=<<table border="0" cellpadding="2" cellspacing="0" cellborder="0">'
        code += '\n <tr><td align="right" port="i1">Global</td></tr>'
        code += '\n <tr><td align="right" port="i2">One Shot</td></tr>'
        code += '\n <tr><td align="right" port="i3">Multiple Shot</td></tr>'
        code += '\n <tr><td align="right" port="i4">I/O</td></tr>'
        code += "\n </table>>]"
        code += '\n key2 [label=<<table border="0" cellpadding="2" cellspacing="0" cellborder="0">'
        code += '\n<tr><td port="i1">&nbsp;</td></tr>'
        code += '\n<tr><td port="i2">&nbsp;</td></tr>'
        code += '\n<tr><td port="i3">&nbsp;</td></tr>'
        code += '\n<tr><td port="i4">&nbsp;</td></tr>'
        code += "\n </table>>]"
        code += '\n  key:i1:e -> key2:i1:w [arrowhead = "odiamond"]'
        code += '\n  key:i2:e -> key2:i2:w [arrowhead = "none"]'
        code += '\n  key:i3:e -> key2:i3:w [style = "dashed", arrowhead = "none"]'
        code += '\n  key:i4:e -> key2:i4:w [arrowhead = "none", color="Green"]'
        code += "\n } \n"

        code += "}\n"

        with open(name_file + ".dot", "w") as file:
            file.write(code)

        print(
            "A graph can be produced with graphviz with one of these commands:\n"
            f"dot {name_file}.dot -Tpng -o {name_file}.png && eog {name_file}.png\n"
            f"dot {name_file}.dot -Tx11"
        )



# import h5py

from fluidimage import SeriesOfArrays
from fluidimage.data_objects.piv import LightPIVResults
from fluidimage.works.piv import WorkPIV

params = WorkPIV.create_default_params()

# # for a very short computation
# params.piv0.shape_crop_im0 = 128
# params.piv0.grid.overlap = 0.

# params.piv0.method_subpix = 'centroid'
# params.piv0.method_correl = 'pythran'

params.multipass.number = 1
params.multipass.use_tps = False
# params.multipass.coeff_zoom = [2, 2]

# bug params.piv0.shape_crop_im0 = 128  # !!
params.piv0.shape_crop_im0 = 64  # (80, 90)
# params.piv0.shape_crop_im1 = (38, 36)
params.fix.correl_min = 0.2
params.fix.threshold_diff_neighbour = 4
# params.piv0.grid.overlap = 10

piv = WorkPIV(params=params)

series = SeriesOfArrays("../../../image_samples/Oseen/Images", "i+1:i+3")
serie = series.get_serie_from_index(0)

result = piv.calcul(serie)

result.display()

result.save()

# lightresult = result.make_light_result()
# lightresult.save()

# lightresultload = LightPIVResults(str_path='piv_Oseen_center01-02_light.h5')

# f=h5netcdf.File('piv_Oseen_center01-02.h5')
# f=h5py.File('piv_Oseen_center01-02.h5')



"""Preprocess toolbox (:mod:`fluidimage.preproc._toolbox_py`)
==========================================================
A toolbox for preprocessing images.
Utilizes functions available from `scipy.ndimage` and `skimage` packages.
cf. http://www.scipy-lectures.org/advanced/image_processing/

Provides:

   :members:

"""

import numpy as np
import scipy.ndimage as ndi

from .io import iterate_multiple_imgs, multiple_imgs_as_ndarray

try:
    import skimage
    from skimage import exposure, filters, morphology

    if skimage.__version__ < "0.13.0":
        print(
            "Warning: to use fluidimage.preproc, "
            "first upgrade scikit-image to a version >= 0.13.0."
        )
except ImportError:
    print(
        "Warning: ImportError, to use fluidimage.preproc, "
        "first install scikit-image >= 0.13.0"
    )


__all__ = [
    "sliding_median",
    "sliding_minima",
    "sliding_percentile",
    "temporal_median",
    "temporal_minima",
    "temporal_percentile",
    "global_threshold",
    "adaptive_threshold",
    "rescale_intensity",
    "equalize_hist_global",
    "equalize_hist_local",
    "equalize_hist_adapt",
    "gamma_correction",
    "sharpen",
    "rescale_intensity_tanh",
]


def imstats(img, hist_bins=256):
    histogram = ndi.measurements.histogram(img, bins=hist_bins)
    return histogram


# ----------------------------------------------------
#   SPATIAL FILTERS
# ----------------------------------------------------


@iterate_multiple_imgs
def sliding_median(
    img=None, weight=1.0, window_size=30, boundary_condition="reflect"
):
    """
    Subtracts the median calculated within a sliding window from the centre of
    the window.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    weight : scalar
        Fraction of median to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0, 1.0).
    window_size : scalar or tuple
        Sets the size of the sliding window.
        Specifying `window_size=3` is equivalent to `window_size=(3,3)`.
    boundary_condition : {'reflect', 'constant', 'nearest', 'mirror', 'wrap'}
        Mode of handling array borders.

    """
    img_out = img - weight * ndi.median_filter(
        img, size=window_size, mode=boundary_condition
    )
    img_out[img_out < 0] = 0
    return img_out


@iterate_multiple_imgs
def sliding_percentile(
    img=None,
    percentile=10.0,
    weight=1.0,
    window_size=30,
    boundary_condition="reflect",
):
    """
    Flexible version of median filter. Low percentile values work well
    for dense images.

    Parameters
    ----------
    img : array_like
        Series of images as a 3D numpy array, or a list or a set
    percentile : scalar
        Percentile to filter. Setting `percentile = 50` is equivalent
        to a `sliding_median` filter.
    weight : scalar
        Fraction of median to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0, 1.0).
    window_shape : tuple of integers
        Specifies the shape of the window as follows (dt, dy, dx)
    boundary_condition : {'reflect', 'constant', 'nearest', 'mirror', 'wrap'}
        Mode of handling array borders.

    """
    img_out = img - weight * ndi.percentile_filter(
        img, percentile, size=window_size, mode=boundary_condition
    )
    img_out[img_out < 0] = 0
    return img_out


@iterate_multiple_imgs
def sliding_minima(
    img=None, weight=1.0, window_size=30, boundary_condition="reflect"
):
    """
    Subtracts the minimum calculated within a sliding window from the centre of
    the window.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    weight : scalar
        Fraction of minima to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0,1.0).
    window_size : scalar or tuple
        Sets the size of the sliding window.
        Specifying `window_size=3` is equivalent to `window_size=(3,3)`.
    boundary_condition : {'reflect', 'constant', 'nearest', 'mirror', 'wrap'}
        Mode of handling array borders.

    """
    img_out = img - weight * ndi.minimum_filter(
        img, size=window_size, mode=boundary_condition
    )
    img_out[img_out < 0] = 0
    return img_out


# ----------------------------------------------------
#   SPATIO-TEMPORAL FILTERS
# ----------------------------------------------------


def _calcul_windowshape(arr_shape, window_shape):
    nb_imgs = arr_shape[0]
    if len(arr_shape) <= 2 or nb_imgs <= 1:
        raise ValueError(
            "Need more than one image to apply temporal filtering "
            "(use sliding filter?)."
        )

    if window_shape is None:
        window_shape = (nb_imgs, 1, 1)
    elif isinstance(window_shape, int):
        window_shape = (window_shape, 1, 1)
    elif not isinstance(window_shape, (tuple, list)):
        raise ValueError("window_shape must be a tuple or a list.")

    elif len(window_shape) == 2:
        window_shape = (nb_imgs,) + window_shape
    elif window_shape[0] <= 1:
        raise ValueError(
            "Cannot perform temporal filtering, try spatial filtering."
        )

    return window_shape


@multiple_imgs_as_ndarray
def temporal_median(img=None, weight=1.0, window_shape=None):
    """
    Subtracts the median calculated in time and space, for each pixel.
    Median filter works well for sparse images.

    Parameters
    ----------
    img : array_like
        Series of images as a 3D numpy array, or a list or a set
    weight : scalar
        Fraction of median to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0,1.0).
    window_shape : tuple of integers
        Specifies the shape of the window as follows (nt, ny, nx) or (ny, nx).

    """
    window_shape = _calcul_windowshape(img.shape, window_shape)

    img_out = img - weight * ndi.median_filter(img, size=window_shape)
    return img_out


@multiple_imgs_as_ndarray
def temporal_percentile(img=None, percentile=10.0, weight=1.0, window_shape=None):
    """
    Flexible version of median filter. Low percentile values work well
    for dense images.

    Parameters
    ----------
    img : array_like
        Series of images as a 3D numpy array, or a list or a set
    percentile : scalar
        Percentile to filter. Setting `percentile = 50` is equivalent
        to a `temporal_median` filter.
    weight : scalar
        Fraction of median to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0,1.0).
    window_shape : tuple of integers
        Specifies the shape of the window as follows (nt, ny, nx) or (ny, nx).

    """
    window_shape = _calcul_windowshape(img.shape, window_shape)

    img_out = img - weight * ndi.percentile_filter(
        img, percentile, size=window_shape
    )
    return img_out


@multiple_imgs_as_ndarray
def temporal_minima(img=None, weight=1.0, window_shape=None):
    """Subtracts the minima calculated in time and space, for each pixel.

    Parameters
    ----------
    imgs : array_like
        Series of images as a 3D numpy array, or a list or a set
    weight : scalar
        Fraction of minima to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0,1.0).
    window_shape : tuple of integers
        Specifies the shape of the window as follows (nt, ny, nx) or (ny, nx).

    """
    window_shape = _calcul_windowshape(img.shape, window_shape)
    img_out = img - weight * ndi.minimum_filter(img, size=window_shape)
    return img_out


# ----------------------------------------------------
#   BRIGHTNESS / CONTRAST TOOLS
# ----------------------------------------------------


@iterate_multiple_imgs
def global_threshold(img=None, minima=0.0, maxima=65535.0):
    """
    Trims pixel intensities which are outside the interval (minima, maxima).

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object

    minima, maxima : float
        Sets the threshold

    """
    img_out = img.copy()
    img_out[img_out < minima] = minima
    img_out[img_out > maxima] = maxima
    return img_out


@iterate_multiple_imgs
def adaptive_threshold(img=None, window_size=5, offset=0):
    """
    Adaptive threshold transforms a grayscale image to a binary image.
    Useful in identifying particles.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    window_size : scalar
        Sets the size of the pixel neighbourhood to calculate threshold.
    offset : scalar
        Constant to be subtracted from the mean.

    """
    img_out = filters.threshold_local(img, window_size, offset=offset)
    return img_out


@iterate_multiple_imgs
def rescale_intensity(img=None, minima=0.0, maxima=4096):
    """
    Rescale image intensities, between the specified minima and maxima,
    by using a multiplicative factor.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    minima, maxima : float
        Sets the range to which current intensities have to be rescaled.

    """
    out_range = (minima, maxima)
    img_out = exposure.rescale_intensity(img, out_range=out_range)
    return img_out


@iterate_multiple_imgs
def rescale_intensity_tanh(img=None, threshold=None):
    """
    Rescale image intensities, using a tanh fit. The maximum intensity of the
    output is set by the threshold parameter.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    threshold:
        Value of intensity with which img is normalized
        img_out = max(img) * tanh( img / threshold)
        If threshold is None: threshold = 2 * np.sqrt(np.mean(img**2))

    """
    if threshold is None:
        threshold = 2 * np.sqrt(np.mean(img**2))

    if threshold == 0:
        return img

    img_out = np.tanh(img / threshold)
    img_out = np.floor(np.max(img) / np.max(img_out) * img_out)
    return img_out


@iterate_multiple_imgs
def equalize_hist_global(img=None, nbins=256):
    """Increases global contrast of the image. Equalized image would have a
    roughly linear cumulative distribution function for each pixel
    neighborhood. It works well when pixel intensities are nearly uniform
    [1,2].

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    nbins : integer
        Number of bins to calculate histogram

    References
    ----------
    - http://scikit-image.org/docs/dev/auto_examples/color_exposure/plot_local_equalize.html
    - https://en.wikipedia.org/wiki/Histogram_equalization

    """
    img_out = exposure.equalize_hist(img, nbins=nbins, mask=None)
    return img_out


@iterate_multiple_imgs
def equalize_hist_adapt(img=None, window_shape=(10, 10), nbins=256):
    """
    Contrast Limited Adaptive Histogram Equalization (CLAHE).
    Increases local contrast.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    window_shape : tuple of integers
        Specifies the shape of the window as follows (dx, dy)
    nbins : integer
        Number of bins to calculate histogram

    References
    ----------
    - http://scikit-image.org/docs/dev/auto_examples/color_exposure/plot_local_equalize.html
    - https://en.wikipedia.org/wiki/Histogram_equalization

    """
    minimum = img.min()
    maximum = img.max()
    img = rescale_intensity(img, 0, 1)
    img = exposure.equalize_adapthist(img, kernel_size=window_shape, nbins=nbins)
    img_out = rescale_intensity(img, minimum, maximum)
    return img_out


@iterate_multiple_imgs
def equalize_hist_local(img=None, radius=10):
    """
    Adaptive histogram equalization (AHE) emphasizes every local graylevel variations [1].
    Caution: It has a tendency to overamplify noise in homogenous regions [2].

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    radius : integer
        Radius of the disk shaped window.

    References
    ----------
    - http://scikit-image.org/docs/dev/auto_examples/color_exposure/plot_local_equalize.html
    - https://en.wikipedia.org/wiki/Adaptive_histogram_equalization

    """
    selem = morphology.disk(radius)
    minimum = img.min()
    maximum = img.max()
    img = rescale_intensity(img, 0, 1)
    img = filters.rank.equalize(img, selem, mask=None)
    img_out = rescale_intensity(img, minimum, maximum)
    return img_out


@iterate_multiple_imgs
def gamma_correction(img=None, gamma=1.0, gain=1.0):
    r"""
    Gamma correction or power law transform. It can be expressed as:

    .. math::
        I_{out} = gain \times {I_{in}} ^ {\gamma}

    Adjusts contrast without changing the shape of the histogram. For the values
    .. :math:`\gamma > 1` : Histogram shifts towards left (darker)
    .. :math:`\gamma < 1` : Histogram shifts towards right (lighter)

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    gamma : float
        Non-negative real number
    gain : float
        Multiplying factor

    References
    ----------
    - http://scikit-image.org/docs/dev/auto_examples/color_exposure/plot_log_gamma.html
    - http://en.wikipedia.org/wiki/Gamma_correction

    """
    img_out = exposure.adjust_gamma(img, gamma, gain)
    return img_out


@iterate_multiple_imgs
def sharpen(img=None, sigma1=3.0, sigma2=1.0, alpha=30.0):
    """
    Sharpen image edges.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    sigma1, sigma2 : float
        Std deviation for two passes gaussian filters. sigma1 > sigma2
    alpha : float
        Factor by which the image will be sharpened

    """
    blurred = ndi.gaussian_filter(img, sigma1)
    filter_blurred = ndi.gaussian_filter(blurred, sigma2)

    img_out = blurred + alpha * (blurred - filter_blurred)
    return img_out



"""Executor async/await using servers (:mod:`fluidimage.executors.exec_async_servers`)
======================================================================================

A executor using async for IO and servers for CPU-bounded tasks.

.. autoclass:: ExecutorAsyncServers
   :members:
   :private-members:

"""

import os
import signal
from pathlib import Path

import numpy as np
import trio

from fluiddyn import time_as_str
from fluidimage.util import log_debug, logger

from .exec_async import ExecutorAsync
from .servers import launch_server

max_items_in_server = 4


class ExecutorAsyncServers(ExecutorAsync):
    """Executor async/await using servers.

    Parameters
    ----------

    nb_max_workers : None, int

      Limits the numbers of workers working in the same time.

    nb_items_queue_max : None, int

      Limits the numbers of items that can be in a output_queue.

    sleep_time : None, float

      Defines the waiting time (from trio.sleep) of a function. Functions await
      "trio.sleep" when they have done a work on an item, and when there is
      nothing in there input_queue.

    """

    _type_server = "multiprocessing"

    def _init_log_path(self):
        name = "_".join(("log", time_as_str(), str(os.getpid())))
        path_dir_log = self.path_dir_exceptions = self.path_dir_result / name
        path_dir_log.mkdir(exist_ok=True)
        self._log_path = path_dir_log / (name + ".txt")

    def __init__(
        self,
        topology,
        path_dir_result,
        nb_max_workers=None,
        nb_items_queue_max=None,
        sleep_time=0.01,
        logging_level="info",
        stop_if_error=False,
    ):
        if stop_if_error:
            raise NotImplementedError

        super().__init__(
            topology,
            path_dir_result,
            nb_max_workers,
            nb_items_queue_max,
            sleep_time=sleep_time,
            logging_level=logging_level,
        )

        # create nb_max_workers servers
        self.workers = []

        for ind_worker in range(self.nb_max_workers):
            log_path = Path(
                str(self._log_path).split(".txt")[0]
                + f"_multi{ind_worker:03}.txt"
            )
            self.workers.append(
                launch_server(
                    topology,
                    log_path,
                    self._type_server,
                    sleep_time,
                    logging_level,
                )
            )

        def signal_handler(sig, frame):
            del sig, frame
            logger.info("Ctrl+C signal received...")

            for worker in self.workers:
                worker.terminate()

            self._has_to_stop = True
            self.nursery.cancel_scope.cancel()
            # we need to raise the exception
            raise KeyboardInterrupt

        signal.signal(signal.SIGINT, signal_handler)

    def compute(self):
        """Compute the whole topology.

        Begin by executing one shot jobs, then execute multiple shots jobs
        implemented as async functions.  Warning, one shot jobs must be ancestors
        of multiple shots jobs in the topology.

        """
        self._init_compute()
        for worker in self.workers:
            worker.send(("__t_start__", self.t_start))
        self.exec_one_shot_works()
        trio.run(self.start_async_works)
        self._finalize_compute()

    async def start_async_works(self):
        """Create a trio nursery and start all async functions."""
        async with trio.open_nursery() as self.nursery:
            for af in reversed(self.async_funcs.values()):
                self.nursery.start_soon(af)

            self.nursery.start_soon(self.update_has_to_stop)

        logger.info("terminate the servers")
        for worker in self.workers:
            worker.terminate()

    async def update_has_to_stop(self):
        """Work has to stop flag. Check if all works has been done.

        Return True if there are no workers in working and if there is no items in
        all queues.

        """

        while not self._has_to_stop:

            result = (
                (not any([bool(queue) for queue in self.topology.queues]))
                and all(worker.is_unoccupied for worker in self.workers)
                and self.nb_working_workers_io == 0
            )

            if result:
                self._has_to_stop = True
                log_debug(f"has_to_stop!")

            if self.logging_level == "debug":
                log_debug(f"self.topology.queues: {self.topology.queues}")
                log_debug(
                    "[worker.is_unoccupied for worker in self.workers]: "
                    f"{[worker.is_unoccupied for worker in self.workers]}"
                )
                log_debug(
                    "[worker.is_available for worker in self.workers]: "
                    f"{[worker.is_available for worker in self.workers]}"
                )
                log_debug(
                    "[worker.nb_items_to_process for worker in self.workers] "
                    f"{[worker.nb_items_to_process for worker in self.workers]}"
                )
                log_debug(
                    f"self.nb_working_workers_io: {self.nb_working_workers_io}"
                )

            await trio.sleep(self.sleep_time)

    async def async_run_work_cpu(self, work, worker):
        """Is destined to be started with a "trio.start_soon".

        Executes the work on an item (key, obj), and add the result on
        work.output_queue.

        Parameters
        ----------

        work :

          A work from the topology

        worker : .servers.WorkerMultiprocessing

          A client to communicate with the server worker.

        """
        try:
            key, obj = work.input_queue.pop_first_item()
        except KeyError:
            worker.is_available = True
            return

        if work.check_exception(key, obj):
            worker.is_available = True
            return

        def run_process():
            # create a communication channel
            parent_conn, child_conn = worker.new_pipe()
            # send (work, key, obj, comm) to the server
            worker.send_job((work.name, key, obj, child_conn))
            worker.is_available = True
            # wait for the end of the computation
            work_name_received, key_received, result = parent_conn.recv()
            assert work.name == work_name_received
            assert key == key_received
            return result

        ret = await trio.to_thread.run_sync(run_process)
        if work.output_queue is not None:
            work.output_queue[key] = ret
        worker.well_done_thanks()

    def def_async_func_work_cpu(self, work):
        async def func(work=work):

            while True:
                while not work.input_queue or (
                    work.output_queue is not None
                    and len(work.output_queue) >= self.nb_items_queue_max
                ):
                    if self._has_to_stop:
                        return
                    await trio.sleep(self.sleep_time)

                available_worker = False
                while not available_worker:
                    if self._has_to_stop:
                        return
                    available_worker = self.get_available_worker()
                    await trio.sleep(self.sleep_time)

                self.nursery.start_soon(
                    self.async_run_work_cpu, work, available_worker
                )
                await trio.sleep(self.sleep_time)

        return func

    def get_available_worker(self):
        """Get a worker available to receive a new job"""
        available_workers = [
            worker
            for worker in self.workers
            if worker.is_available
            and worker.nb_items_to_process < max_items_in_server
        ]

        if not available_workers:
            return False

        nb_items_to_process_workers = [
            worker.nb_items_to_process for worker in available_workers
        ]

        index = np.argmin(nb_items_to_process_workers)
        worker = available_workers[index]
        worker.is_available = False
        return worker


class ExecutorAsyncServersThreading(ExecutorAsyncServers):
    """Just used to get a better coverage"""

    _type_server = "threading"



"""2D/3D Calibration using OpenCV (:mod:`fluidimage.calibration.calib_cv`)
==========================================================================

.. autoclass:: ParamContainerCV
   :members:
   :private-members:

.. autoclass:: SimpleCircleGrid
   :members:
   :private-members:

.. autoclass:: CalibCV
   :members:
   :private-members:

"""
import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle
from matplotlib.transforms import Bbox
from scipy.interpolate import griddata

from fluiddyn.util.paramcontainer import ParamContainer
from fluidimage._opencv import cv2


class ParamContainerCV(ParamContainer):
    """A container to easily display OpenCV parameters"""

    def __init__(self, params_cv, tag="OpenCV"):
        super().__init__(tag)
        self._set_internal_attr("_params_cv", params_cv)

        attrs = {
            a: getattr(params_cv, a)
            for a in dir(params_cv)
            if not callable(a) and not a.startswith("__")
        }
        for key, value in attrs.items():
            self._set_attrib(key, value)

    def _as_params_cv(self):
        params_cv = self["_params_cv"]
        for key in self._get_key_attribs():
            value = self[key]
            setattr(params_cv, key, value)
        return params_cv


class SimpleCircleGrid:
    """Detect the centers of calibration target consisting of a grid of circle
    shaped points. Use it to detect the image points.

    """

    @staticmethod
    def create_default_params():
        params_cv = cv2.SimpleBlobDetector_Params()
        # Slightly nicer defaults
        params_cv.filterByColor = False
        params_cv.minArea = 0.0
        return ParamContainerCV(params_cv, "SimpleBlobDetector")

    def __init__(self, params: ParamContainerCV) -> None:
        self.params = params
        params_cv = params._as_params_cv()
        self.detector = cv2.SimpleBlobDetector_create(params_cv)

    def detect_all(self, image: np.ndarray, debug=False):
        """Detects all blobs as per parameters without any constraints.

        Parameters
        ----------
        image: array
            A calibration image of maximum intensity 255.
        debug : bool
            Plot the detected points and the bounding box

        """
        image = image.astype(np.uint8)
        keypoints = self.detector.detect(image)
        if debug:
            plt.figure()
            for k in keypoints:
                plt.scatter(*k.pt)
            plt.show()

        if len(keypoints) == 0:
            raise ValueError("No blob detected")
        return keypoints

    def detect_grid(
        self,
        image: np.ndarray,
        origin: tuple,
        nx: int,
        ny: int,
        ds: float,
        debug=False,
    ):
        """Detect a ``nx`` by ``ny`` circle grid centered around an origin.

        Parameters
        ----------
        image: array
            A calibration image of maximum intensity 255.
        origin: tuple
            Origin / Principal point location in pixel coordinates
        nx, ny : int
            Shape of the grid, i.e. number of points
        ds : float
            Grid spacing in pixel coordinates
        debug : bool
            Plot the detected points and the bounding box

        """
        keypoints = self.detect_all(image)
        # Add 1 so that the points at the edge of the bounding box are included
        w = (nx + 1) * ds
        h = (ny + 1) * ds
        originx, originy = origin
        bbox = Bbox.from_bounds(originx - w // 2, originy - h // 2, w, h)
        if debug:
            print(bbox)

        center_list = []
        for k in keypoints:
            if bbox.contains(*k.pt):
                center_list.append(k.pt)

        centers = np.array(center_list)
        xround = np.round(centers[..., 0] / ds)
        yround = np.round(centers[..., 1] / ds)
        # Sort as a grid
        ind = np.lexsort((xround, yround))
        # Add another dimension
        centers = np.array([centers[ind]], dtype=np.float32)

        if len(centers[0]) != nx * ny:
            raise AssertionError(f"Only {len(centers[0])} points were found")

        if debug:
            fig, ax = plt.subplots()
            ax.imshow(image, cmap="gray")
            ax.scatter(centers[..., 0], centers[..., 1])
            l, b, w, h = bbox.bounds
            ax.add_patch(
                Rectangle(xy=(l, b), width=w, height=h, edgecolor="r", fill=False)
            )

        return centers


def construct_object_points(nx: int, ny: int, z: float, ds: float):
    """Prepare object points in world coordinates, as flattened list of
    coordinates such as::

        (0,0,z), (1,0,z), (2,0,z) ....,(6,5,z)

    This format is expected by OpenCV's ``calibrateCamera`` function.

    Parameters
    ----------
    nx, ny : int
        Shape of the grid, i.e. number of points
    z : float
        z-location in world coordinates
    ds : float
        Grid spacing in world coordinates

    """
    objp = np.zeros((nx * ny, 3), np.float32)
    assert nx % 2 == 1
    assert ny % 2 == 1
    stepx, stepy = np.array((nx, ny), dtype=int) * 1j
    # nx, ny will refer the number of points on one side from now on
    nx = (nx - 1) // 2
    ny = (ny - 1) // 2
    objp[:, :2] = np.mgrid[-ny:ny:stepy, -nx:nx:stepx].T.reshape(-1, 2) * ds
    objp[:, 2] = z
    return objp


class CalibCV:
    """Calibrate a camera and save them as HDF5 files. Also use this to load
    saved calibrations and interpolate extrinsic parameters (rotation and
    translation) while reconstructing.

    """

    def __init__(self, path_file="cam.h5"):
        self.path_file = str(path_file)
        if os.path.exists(path_file):
            print(f"Loading {path_file}.")
            self.params = ParamContainer(path_file=self.path_file)

    def save(self, zs, ret, mtx, dist, rvecs, tvecs):
        self.params = params = ParamContainer(tag="CalibCV")
        params._set_attribs(
            {
                "class": self.__class__.__name__,
                "module": self.__module__,
                "f": np.diag(mtx)[:2],
                "C": mtx[0:2, 2],
                "cam_mtx": mtx,
                "kc": dist.T[0],
                "rotation": np.array(rvecs),
                "translate": tvecs,
                "zs": np.array(zs),
            }
        )
        path_dir = os.path.dirname(self.path_file)
        os.makedirs(path_dir, exist_ok=True)
        if os.path.exists(self.path_file):
            print(f"WARNING: {self.path_file} already exists. Skipping save.")
        else:
            params._save_as_hdf5(self.path_file)

    def rotmtx_from_rotvec(self, rot_vec):
        rot_mtx, rot_jac = cv2.Rodrigues(rot_vec)
        return rot_mtx

    def get_rotation(self, znew):
        """Linearly interpolate the rotation vector based on z location."""
        rot_vec = griddata(self.params.zs, self.params.rotation, znew)
        return rot_vec

    def get_translate(self, znew):
        """Linearly interpolate the translation vector based on z location."""
        translate = griddata(self.params.zs, self.params.translate, znew)
        return translate

    def calibrate(
        self,
        imgpoints: list,
        objpoints: list,
        zs: list,
        im_shape: tuple,
        origin=None,
        debug=False,
        flags=None,
    ):
        """Calibrate a camera based on a list of image points (in pixel
        coordinates) and object points (in world coordinates) and the z-locations
        (in world coordinates).

        Parameters
        ----------
        imgpoints : list of arrays
            Image points in pixel coordinates. Use `SimpleCircleGrid` to detect
            them from a single calibration image. Append such image points
            from multiple calibration images in a list.
        objpoints : list of arrays
            Object points as produced by function `construct_object_points`
            constitute a single array. Likewise append them into a list for
            multiple calibration images.
        zs : list of float
            List of z locations of the calibration targes in world coordinates.
        im_shape : tuple of int
            Image dimensions in pixels.
        origin : tuple, optional
            Origin / Principal point location in pixel coordinates
        flags : int, optional
            OpenCV specific calibration flags
        debug : bool, optional
            Return the result if true, else save it as an XML file.

        """
        # Initial guesses
        initial_mtx = np.eye(3)
        if origin is not None:
            # Origin / Principal point at center:
            initial_mtx[0:2, 2] = origin
        initial_dist = np.zeros((5, 1))

        if flags is None:
            flags = (
                cv2.CALIB_USE_INTRINSIC_GUESS
                + cv2.CALIB_FIX_K4
                + cv2.CALIB_FIX_K5
            )

        result = cv2.calibrateCamera(
            objpoints, imgpoints, im_shape, initial_mtx, initial_dist, flags=flags
        )

        if debug:
            return result
        else:
            self.save(zs, *result)



#!/usr/bin/env python
"""PyQtGraph app which opens a single image or multiple images
in separate docks / as a slideshow.

Examples
--------
$ fluidimviewer-pg im1.png im2.png
$ fluidimviewer-pg im???a.png
$ fluidimviewer-pg -s im???a.png

Dock mode
---------
* Displays histogram of image data with movable region defining the dark/light
  levels
* Region of interest (ROI) and embedded plot for measuring image values
  across frames
* Image normalization / background subtraction

Slideshow mode: Keyboard interaction
------------------------------------
* left/right arrows step forward/backward 1 frame when pressed, seek at 20fps
  when held.
* up/down arrows seek at 100fps
* pgup/pgdn seek at 1000fps
* home/end seek immediately to the first/last frame
* space begins playing frames. If time values (in seconds) are given for each
  frame, then playback is in realtime.

"""
import argparse
import os
import sys

from fluidimage import config_logging
from fluidimage.gui.pg_wrapper import PGWrapper


def dock(args, for_testing):
    pg = PGWrapper()
    for arg in args:
        title = os.path.basename(arg)
        pg._add_dock(title, size=(500, 500), position="right")
        pg.view(arg, title)

    if not for_testing:
        pg.show()
    return pg


def slideshow(args, for_testing):
    title = "FluidImage {} to {}".format(
        os.path.basename(args[0]), os.path.basename(args[-1])
    )
    pg = PGWrapper(title=title)

    pg.view(args, title)
    if not for_testing:
        pg.show()
    return pg


def parse_args(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument("images", help="Path to the image or pattern", nargs="+")

    parser.add_argument(
        "-s",
        "--slideshow",
        help="View multiple images as a slideshow. If not open them in docks.",
        action="store_true",
    )

    parser.add_argument("-v", "--verbose", action="store_true")

    return parser.parse_args(args)


def main(args=None, for_testing=False):
    """Parse arguments and execute `fluidimviewer-pg`."""
    if args is None:
        args = parse_args()

    if args.verbose:
        config_logging("debug")
    else:
        config_logging()

    if args.slideshow:
        return slideshow(args.images, for_testing)
    else:
        return dock(args.images, for_testing)


if __name__ == "__main__":
    main()



"""Topology launcher Qt GUI (:mod:`fluidimage.gui.launcher.main`)

.. autoclass:: Program
   :members:
   :private-members:

"""

import sys

from matplotlib.backends.qt_compat import QtWidgets

from fluiddyn.util.paramcontainer import ParamContainer
from fluiddyn.util.paramcontainer_gui import QtParamContainer
from fluidimage.topologies.launcher import (
    TopologyPIVLauncher,
    TopologyPreprocLauncher,
)

from .mainwindow import Ui_MainWindow


class Program(QtWidgets.QMainWindow, Ui_MainWindow):
    def __init__(self):
        topologies = [TopologyPreprocLauncher, TopologyPIVLauncher]
        self.topologies = {cls.__name__: cls for cls in topologies}

        QtWidgets.QMainWindow.__init__(self)
        self.setupUi(self)

        topo_names = ["TopologyPreprocLauncher", "TopologyPIVLauncher"]
        self.actions = {}

        for topo_name in topo_names:
            action = self.actions[topo_name] = QtWidgets.QAction(self)
            self.menuTopologies.addAction(action)
            action.setText(topo_name)

            def func(_, topo_name=topo_name):
                self.init_topo(topo_name)

            action.triggered.connect(func)

        self.actionOpen.triggered.connect(self.open_file)

    def closeEvent(self, QCloseEvent):
        pass

    def open_file(self):
        path_file = QtWidgets.QFileDialog.getOpenFileName(
            self.centralwidget, "OpenFile"
        )

        params = ParamContainer(path_file=path_file)
        print(params)
        raise NotImplementedError

    def init_topo(self, topo_name):
        Topology = self.topologies[topo_name]
        params = Topology.create_default_params()

        self.qt_params = QtParamContainer(
            params, top=True, module_run_from_xml="fluidimage.run_from_xml"
        )

        # first remove all widgets from verticalLayout_2
        layout = self.verticalLayout_2
        for i in reversed(range(layout.count())):
            widgetToRemove = layout.itemAt(i).widget()
            # remove it from the layout list
            layout.removeWidget(widgetToRemove)
            # remove it from the gui
            widgetToRemove.setParent(None)

        self.verticalLayout_2.addWidget(self.qt_params.page_main)


def main():
    app = QtWidgets.QApplication(sys.argv)
    w = Program()
    w.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""Preprocess base (:mod:`fluidimage.preproc.base`)
===================================================

To preprocess series of images.

Provides:

.. autoclass:: PreprocBase
   :members:
   :private-members:

"""
import os
import sys

from fluidimage.data_objects.display_pre import DisplayPreProc

from .. import ParamContainer, SerieOfArraysFromFiles


class PreprocBase:
    """Preprocess series of images with various tools."""

    @classmethod
    def create_default_params(cls, backend="python"):
        """Class method returning the default parameters.

        Parameters
        ----------

        backend: {'python', 'opencv'}

            Specifies which backend to use.

        """
        params = ParamContainer(tag="params")
        params._set_child("preproc")
        params.preproc._set_child("series", attribs={"path": ""})
        params.preproc.series._set_doc(
            """
Parameters indicating the input series of images.

path : str, {''}

    String indicating the input images (can be a full path towards an image
    file or a string given to `glob`).
"""
        )
        if backend == "python":
            from .toolbox import PreprocToolsPy

            cls._Tools = PreprocToolsPy
        elif backend == "opencv":
            from .toolbox import PreprocToolsCV

            cls._Tools = PreprocToolsCV
        else:
            raise ImportError("Unknown backend: %s" % backend)

        cls._Tools.create_default_params(params)
        return params

    def __init__(self, params=None):
        """Set path for results and loads images as SerieOfArraysFromFiles."""
        if params is None:
            params = self.__class__.create_default_params()

        self.params = params.preproc

        path = params.preproc.series.path
        if not os.path.exists(path):
            path = params.preproc.series.path = os.path.expandvars(path)

        self.serie_arrays = SerieOfArraysFromFiles(path)
        self.tools = self._Tools(params)
        self.results = {}

    def __call__(self):
        """Apply all enabled preprocessing tools on the series of arrays
        and saves them in self.results.

        """
        name_files = self.serie_arrays.get_name_files()
        for i, img in enumerate(self.serie_arrays.iter_arrays()):
            name = name_files[i]
            self.results[name] = self.tools(img)

    def display(self, ind=None, hist=False, results=None):
        nimages = 2
        if not ind:
            name_files = self.serie_arrays.get_name_files()[:nimages]
        else:
            name_files = self.serie_arrays.get_name_files()[ind : ind + nimages]

        before = {}
        for fname in name_files:
            before[fname] = self.serie_arrays.get_array_from_name(fname)

        if results is None:
            results = self.results

        return DisplayPreProc(
            before[name_files[0]],
            before[name_files[1]],
            results[name_files[0]],
            results[name_files[1]],
            hist=hist,
        )


def _make_doc_with_filtered_params_doc(cls):
    params = cls.create_default_params()
    strings = ("Parameters", "References", "----------")
    return "\n".join(
        line
        for line in params._get_formatted_docs().split("\n")
        if not any(line.endswith(string) for string in strings)
    )


if "sphinx" in sys.modules:
    __doc__ += _make_doc_with_filtered_params_doc(PreprocBase)
